from pandas import concat

from ...fold import ns_path
from ...manifest.namings import alias_df
from ..io import mmd

__all__ = ["routing_df"]

assert (routing_mmd := ns_path / "routing.mmd").exists(), f"No file {routing_mmd=}"

pk = ["domain", "route"]  # route is 'dev' or 'local' relative filesystem path
cfg = {"sep": "=", "headersep": False, "labels": pk, "part_keys": pk}
r = mmd(routing_mmd, listparseconfig=cfg)
routing_df = {l.header.contents: l.as_df(forbid_header=True) for l in r.lists}
for routing_ns in routing_df:
    routing_df.get(routing_ns).insert(0, "namespace_full", routing_ns)
routing_df = (
    concat([routing_df.get(ns) for ns in routing_df]).reset_index().drop("index", 1)
)
routing_df.namespace_full = routing_df.namespace_full.astype("category")
routing_df = alias_df.merge(routing_df, on=["namespace_full", "domain"])



from enum import Enum
from os import sep as os_sep
from pathlib import Path

from parse import parse as fparse
from yaml import dump

SubclassablePathType = type(Path())


def check_yaml(yaml_dict, keys):
    "Validate keys of the YAML dict for a GitLab Pages CI config."
    for k in keys:
        assert k in yaml_dict, f"YAML missing key: '{k}'"


class Stage(Enum):
    "There are multiple possible stages but I think GLP just uses 'deploy'."
    # Build = "build"
    # Test = "test"
    Deploy = "deploy"


class Cmd(Enum):
    """
    Fixed strings to validate against (as an enum so as to be immutable and named),
    representing commands in a build script.  The `Cmd.Cp` command is templated,
    using "{}" which will be parsed by `parse` from the package `parse`, described
    as 'the opposite of format()' i.e. like f-strings in reverse.
    """

    Mk = "mkdir .public"
    Cp = "cp -r {}* .public"
    Mv = "mv .public public"
    Cu = "curl -X POST -F token={} -F ref={} https://gitlab.com/api/v4/projects/{}/trigger/pipeline"


class TemplateVarMixIn:
    def set_configured_var_name(self, var_name):
        self._configured_var_name = var_name

    def set_configured_value(self, val):
        self._configured_val = val

    @staticmethod
    def configure_var_str(configured_var):
        "Mask `None` values as the empty string (N.B. also will affect false-y values)"
        configured_str = f"{configured_var}" if configured_var else ""
        return configured_str

    @property
    def as_str(self):
        configured_var = self._configured_val
        configured_str = self.configure_var_str(configured_var)
        return configured_str


class NewTemplateVarMixIn(TemplateVarMixIn):
    def __new__(cls, val):
        return type(super()).__new__(cls, val)


class StrTV(str, NewTemplateVarMixIn):
    pass


class PathTV(SubclassablePathType, NewTemplateVarMixIn):
    @staticmethod
    def configure_var_str(configured_var):
        """
        Must take care not to accidentally let an empty path become the root path
        while ensuring that paths are otherwise always terminated by a path separator
        """
        configured_str = f"{configured_var}{os_sep}" if configured_var else ""
        return configured_str


class FmtCmd(dict):
    """
    A formatted template will take a string such as:

      "Hello {}, the date today is {} the {} of {}, {}."

    along with 1 or more [in this case 5] variable parts, each of which
    has a name, indicated by the keys of a dict. In this case it'd be:

      ["name", "day_of_week", "day_of_month", "month", "year"]

    The values corresponding to these keys in the dict would be `StrTV`
    objects, initialised with 2 pieces of info: the templated string and the
    callable to be applied after it has been parsed (or alternatively `None`).
    """

    def __init__(self, cmd, d):
        self._cmd = cmd
        self._key, *self._keys = d
        if self._keys:
            vars(self)["_keys"].insert(0, self._key)
        vars(self).update({f"_key{'s' if not self._keys else ''}": None})
        # So either (_key = None, _keys = list) OR (_key = string, _keys = None)
        super().__init__(d)

    @property
    def formatted_cmd(self):
        cmd = self._cmd.value  # template string (from the Cmd enum)
        vals = self.configured_values.values()  # set by Script.parse on __init__
        # vals is a list of either: None (no match), or TemplateVarMixIn-inheritor
        # strings = [v.as_str if v else "" for v in vals]
        strings = []
        formatted = cmd.format(*[v.as_str if v else "" for v in vals])
        return formatted

    def set_configured_values(self, val_list):
        for k, val in zip(self, val_list):
            if val:
                template_var = val
                template_var.set_configured_var_name(
                    k
                )  # not needed, just for debugging
                template_var.set_configured_value(val)
        self._configured_values = dict(
            zip(self, val_list)
        )  # {var_name:TemplateVarMixIn}

    @property
    def configured_values(self):
        return self._configured_values


class Script:
    """
    The script (list of commands) to run, relating to serving files from a
    single directory (which must be named 'public').
    """

    def __init__(self, cmd_list, pages=True):
        self.cfg = {}
        self.templates = self.pages_templates if pages else self.trigger_templates
        self.parsed_templates = []
        self.parse(cmd_list)

    pages_templates = {
        Cmd.Mk: None,
        Cmd.Cp: {"subdirectory": PathTV},
        Cmd.Mv: None,
    }
    trigger_templates = {
        Cmd.Cu: ["token", "ref", "project_id"],  # will default to TemplateVar callback
    }

    def parse(self, commands, default_callback=StrTV):
        for i, command in enumerate(commands):
            template_cmd = [*self.templates][i]
            d = self.templates.get(template_cmd)  # indicates how to interpret template
            if d:
                # Extract the variable command from template_cmd and parse against
                # command (i.e. the full string from file) applying any callback
                if type(d) is list:
                    # Check if type(d) is list, apply default callback if so
                    d = {k: default_callback for k in d}
                    # Change it so you can reuse it later (in `as_list`) as a dict
                    self.templates[template_cmd] = d  # template_cmd used as dict key
                fmt = FmtCmd(template_cmd, d)  # store template_cmd as attrib on d
                self.parse_formatted_template(fmt, command)
                fmt.set_configured_values([self.cfg.get(k) for k in fmt])
                self.parsed_templates.append(fmt)
            else:
                # Fixed templated string (no variable part to process), just validate it
                match_str = f"{template_cmd.value=} == {command=}"
                assert template_cmd.value == command, f"Match failed: {match_str}"
                self.parsed_templates.append(template_cmd)

    def parse_formatted_template(self, fmt, command):
        template_cmd = fmt._cmd
        configured_value = fparse(template_cmd.value, command)
        if configured_value:
            vals = configured_value.fixed
        else:
            if len(fmt) == 1:
                vals = [None]
            else:
                e_msg = (
                    "Cannot distinguish empty values for a command template"
                    " with multiple variable parts.\nTODO: handle if needed, i.e. if "
                    "multiple parts can vary and it is valid for them to potentially be "
                    "empty (for now it is invalid for any of the multi-parts to be blank)"
                )
                raise NotImplementedError(e_msg)
                # Previously the code below ran, but this cannot distinguish one from
                # multiple empty/missing values, and will 'blank' the provided ones too
                vals = [None for k in fmt]
        for val, (k, callback) in zip(vals, fmt.items()):
            assert k not in self.cfg, "Key {k} already seen: aborting! Oh no!"
            if val and callable(callback):
                val = callback(val)  # call the supplied function on the parsed value
            self.cfg.update({k: val})

    @property
    def as_list(self):
        cmd_list = []
        for t in self.parsed_templates:
            if isinstance(t, FmtCmd):
                cmd_list.append(t.formatted_cmd)
            elif isinstance(t, Cmd):
                # Fixed template string, just pass its [string] value
                cmd_list.append(t.value)
            else:
                raise TypeError(f"Unexpected command '{t=}' in parsed_templates")
        return cmd_list

    def __repr__(self):
        if not (subdir := self.cfg.get("subdirectory")):
            subdir = ""
        script_repr = " ".join(f"{k}='{v}'" for k, v in self.cfg.items())
        return f"Script <{script_repr}>"


class Artifacts:
    """
    Build artifacts means things not tracked by git to be attached,
    i.e. the 'public' directory created during the Pages build.
    """

    def __init__(self, yaml_dict):
        assert "paths" in yaml_dict, "No paths artifact"
        paths = ["public"]
        self.paths = paths
        assert yaml_dict.get("paths") == paths, f"No artifact {paths=}"

    @property
    def as_dict(self):
        return {"paths": self.paths}

    def __repr__(self):
        cls = self.__class__.__name__
        paths = self.paths
        return f"{cls} <{paths=}>"


class Only:
    "Only run on the given `branch` (default 'master')."

    def __init__(self, yaml_entry, branch="master"):
        assert yaml_entry == [branch], f"No path artifact {branch}"
        self.branch = [branch]

    def __repr__(self):
        cls = self.__class__.__name__
        branch = self.branch
        return f"{cls} <{branch=}>"


class PagesJob:
    "Required to build GitLab Pages: validates YAML and stores as object"

    def __init__(self, yaml_dict, branch="www"):
        check_yaml(yaml_dict, ["stage", "script", "artifacts", "only"])
        self.stage = Stage(yaml_dict.get("stage"))
        self.script = Script(yaml_dict.get("script"), pages=True)
        self.artifacts = Artifacts(yaml_dict.get("artifacts"))
        self.only = Only(yaml_dict.get("only"), branch=branch)

    @property
    def as_dict(self):
        d = {
            "stage": self.stage.value,
            "script": self.script.as_list,
            "artifacts": self.artifacts.as_dict,
            "only": self.only.branch,
        }
        return d

    def __repr__(self):
        cls = self.__class__.__name__
        return f"{cls}: {{{self.stage}, {self.script}, {self.artifacts}, {self.only}}}"


class TriggerQuillJob:
    """
    Required to push the built site to the www branch via quill repo CI: validates YAML
    and stores as object
    """

    def __init__(self, yaml_dict, branch="master"):
        check_yaml(yaml_dict, ["stage", "script", "only"])
        self.stage = Stage(yaml_dict.get("stage"))
        self.script = Script(yaml_dict.get("script"), pages=False)
        self.only = Only(yaml_dict.get("only"), branch=branch)

    @property
    def as_dict(self):
        d = {
            "stage": self.stage.value,
            "script": self.script.as_list,
            "only": self.only.branch,
        }
        return d

    def __repr__(self):
        cls = self.__class__.__name__
        return f"{cls}: {{{self.stage}, {self.script}, {self.only}}}"


class SiteCI:
    "Representing (and validating) a `gitlab-ci.yml` config deployed to GitLab Pages."

    def __init__(self, yaml_dict):
        check_yaml(yaml_dict, ["trigger_quill", "pages"])
        self.quill_trigger = TriggerQuillJob(yaml_dict.get("trigger_quill"))
        self.pages = PagesJob(yaml_dict.get("pages"))

    @property
    def pages(self):
        return self._pages

    @pages.setter
    def pages(self, pages):
        self._pages = pages

    def __repr__(self):
        cls = self.__class__.__name__
        return f"{cls}: {self.pages}"

    @property
    def as_dict(self):
        return {
            "trigger_quill": self.quill_trigger.as_dict,
            "pages": self.pages.as_dict,
        }

    @property
    def as_yaml(self):
        return dump(self.as_dict, sort_keys=False)



from .elems import BaseElem

__all__ = ["DocElem", "DocLists"]


class DocElem(BaseElem):
    """
    Base class for any doc-level element. Initialised as either
    a list of `Node` objects (stored as the `nodes` attribute) or
    a list of objects which themselves have a `nodes` attribute
    (stored as the `items` attribute).
    """

    def __init__(self, nodes=None, nodelists=None):
        super().__init__(nodes, nodelists)


class DocLists(DocElem, list):
    """
    A doc-level 'holder' for block-level [by definition] lists,
    providing access to the union (over all blocks in the Doc)
    of blocks' lines' output from parsing the block nodes to
    lists (however there is no need to confine the lists to
    block-level features after they have been parsed as such).
    """

    # TODO: wasteful to store nodes twice: just store index for header/list items
    # i.e. some 'view' on nodes, not a duplicate copy of the nodes
    def __init__(self, nodelists):
        super().__init__()
        self.extend(nodelists)

    @property
    def repr_str(self):
        n_lists = len(self)
        s = "s" if n_lists > 1 else ""
        return f"{n_lists} list{s}"


# Example use case
class DocHeaderTree(DocElem):
    """
    A doc-level header tree could be a hierarchy of header nodes.
    """

    def __init__(self, nodelist):
        super().__init__(nodelist)



from configparser import ConfigParser
from pathlib import Path
from sys import stderr

from ..__share__ import ql_path
from .distrib_setup import create_manifest, create_ns_aliases, create_routing_table
from .subdomain import parse_subdomain_url

__all__ = ["ns", "ns_path"]


def resolve_ns_path(qp=ql_path, ini_fn="spin.ini"):
    c = ConfigParser()
    with open(qp / ini_fn, "r") as f:
        c.read_file(f, ini_fn)
    rel_level = c.get("PARENT", "relative")
    dir_name = c.get("PARENT", "dir_name")
    ### This part is all about standardising whether local or distributed
    home_spin = Path.home() / "spin"
    home_is_setup = (qp / Path(*[".."] * 3)).resolve() == home_spin
    if home_is_setup:
        ns_p = (qp / rel_level / dir_name).resolve()
        pre_existing_ns_p = True
    else:
        # this is not set up in the home directory, probably been distributed
        # if you resolve the namespace path you'll only check if the library
        # "ss" has been installed, which on PyPi is a subtitles library
        # (Don't do this! Instead set up an installation in the home directory)
        ns_p = home_spin / dir_name
        pre_existing_ns_p = ns_p.exists()
        if not pre_existing_ns_p:
            ns_p.mkdir(parents=True)
            create_ns_aliases(ns_p)
            create_routing_table(ns_p)
            create_manifest(ns_p)
    return ns_p, pre_existing_ns_p


def read_ns(qp=ql_path, ini_fn="spin.ini"):
    ns_p, pre_existing_ns_p = resolve_ns_path(qp, ini_fn)
    ns_subdirs = [d.name for d in ns_p.iterdir() if d.is_dir()]
    ns = {k: parse_subdomain_url(ns_p, k) for k in ns_subdirs}
    return ns, pre_existing_ns_p


class NameSpaceDict(dict):
    def __init__(self):
        self.refresh()

    def refresh(self):
        fresh, pre_existing_ns_p = read_ns()
        if self != fresh:
            self.clear()
            self.update(fresh)


ns = NameSpaceDict()  # NameSpace()
ns_path, pre_existing_ns_p = resolve_ns_path()



from ..fold import ns_path as ss
from ..scan import mmd
from ..scan.lever.lists import SepBlockList
from .parsing import read_man

__all__ = ["ssm"]

ssm_p = ss / "manifest.mmd"
pk = ("domain", "repo_name", "branches")
sep_config = {"sep": ":", "headersep": True, "labels": None}  # don't need attrs
config_dict = {"listclass": SepBlockList, "part_keys": pk, "listconfig": sep_config}
ssm = mmd(ssm_p, listparseconfig=config_dict)
ssm.repos = property(read_man).__get__(ssm)



class BaseElem:
    "Base class for any element (both block-level and doc-level)."

    def __init__(self, nodes=None, nodelists=None):
        if nodes:
            self.nodes = nodes
        elif nodelists:
            self.items = nodelists



from .parser import Doc

__all__ = ["MMD"]


class MMD(Doc):
    def __init__(self, mmd_lines, listparseconfig=None):
        super().__init__(mmd_lines, listparseconfig=listparseconfig)

    def __repr__(self):
        return f"Parsed MMD file ({self._doc_repr})"



from git import Repo

from ...fold.ns_util import ns_path


def _has_clean_wt(domain, add_before_check=True):
    """
    Helper function (TODO: move to a dedicated git module)
    """
    repo_dir = ns_path / domain
    assert repo_dir.exists(), f"{repo_dir=!s} in namespace but doesn't exist"
    repo = Repo(repo_dir)
    if add_before_check:
        repo.git.add("--all")
    is_clean = not repo.is_dirty()
    return is_clean


def _active_branch(domain):
    "Return the string name of the active branch of the repo."
    repo = Repo(ns_path / domain)
    active_branch = repo.active_branch.name
    return active_branch



from configparser import ConfigParser


def git_to_https(uri):
    "Convert git URI (for a GitHub SSH repo) to https"
    assert uri.startswith("git"), f"No git SSH repo at {uri}"
    https = uri.split("@")[1].replace(":", "/")
    https = "https://" + https[: https.rfind(".git")]
    return https


def parse_subdomain_url(ns_path, sd_subpath):
    """
    Process an individual subdomain for quill.fold.read_ns
    """
    subdomain = ns_path / sd_subpath
    sd_git_conf = subdomain / ".git" / "config"
    c = ConfigParser()
    with open(sd_git_conf, "r") as f:
        c.read_file(f)
    sd_url = git_to_https(c.get('remote "origin"', "url"))
    return sd_url



from collections.abc import Sequence
from pathlib import Path

from bs4 import BeautifulSoup, Tag

from .html_elem_util import MetaTag
from .transforms import HtmlBlock


class HtmlPage:
    universal_stylesheet_href = "https://spin.systems/stylesheets/sakura-earthly.css"

    def __init__(self, content, depth_from_root, head_params={}):
        self.preamble = "<!doctype html>\n"
        self.content = content
        self.depth = depth_from_root  # this is duplicated in BaseIndexPage, oops...
        self.head = self.make_head(**head_params)

    def as_str(self):
        soup = BeautifulSoup(features="html5lib")
        html = Tag(soup, name="html", attrs={"lang": "en"})
        html.append(self.head)
        body = Tag(name="body")
        c_tags = self.content if isinstance(self.content, Sequence) else [self.content]
        for c in c_tags:
            body.append(c)
        html.append(body)
        return self.preamble + str(html)  # .prettify()

    @property
    def _default_css_param(self):
        css_filenames = ["styles"]
        return css_filenames

    def make_head(self, css=None, desc=None, og_img=None, tw_card=True, sakura=True):
        head = Tag(name="head")
        ct_a = {"http-equiv": "Content-Type", "content": "text/html; charset=utf-8"}
        head.append(MetaTag(ct_a))
        head.append(MetaTag({"charset": "UTF-8"}))
        vp_a = {"name": "viewport", "content": "width=device-width, initial-scale=1"}
        head.append(MetaTag(vp_a))
        if desc:
            head.append(MetaTag({"description": desc}))
        if og_img:
            head.append(MetaTag({"og:image": og_img}))
        if tw_card:
            tw_card_attrs = {"name": "twitter:card", "content": "summary"}
            head.append(MetaTag(tw_card_attrs))
        if css is None:
            css = self._default_css_param
        # Add Sakura to all pages
        css_hrefs = [self.universal_stylesheet_href] if sakura else []
        for css_filename in css:
            css_hrefs.append(
                str(Path(*[".."] * self.depth) / "css" / f"{css_filename}.css")
            )
            # all other stylesheets live in the css directory under the site root
        for css_href in css_hrefs:
            css_attrs = {"rel": "stylesheet", "href": css_href}
            style = Tag(name="link", attrs=css_attrs, can_be_empty_element=True)
            head.append(style)
        return head


class HtmlDoc(HtmlPage):
    def __init__(self, doc, depth_from_root, qa_pair=False, nested_qa_summary=False):
        self.preamble = "<!doctype html>\n"
        self.doc = doc
        self.depth = depth_from_root
        self.qa_pair = qa_pair
        self.nested_qa_summary = nested_qa_summary
        self.head = self.make_head()

    def as_str(self):
        soup = BeautifulSoup(features="html5lib")
        html = Tag(soup, name="html")
        html.append(self.head)
        body = Tag(name="body")
        article = self.process_article()
        body.append(article)
        html.append(body)
        return self.preamble + str(html)  # .prettify()

    def process_article(self):
        article = Tag(name="article")
        unassigned_lists = self.doc.lists.copy()
        for b in self.doc.blocks:
            assignable_lists = []
            for l in unassigned_lists:
                first_list_line = l.all_nodes[0].line_no
                last_block_line = b.end_line
                if first_list_line < last_block_line:
                    # the list is in this block
                    assignable_lists.append(l)
                # do stuff with the assignable lists in this block
            block_html = HtmlBlock(
                b, assignable_lists, self.qa_pair, self.nested_qa_summary
            ).as_soup()
            article.append(block_html)
            for l in assignable_lists:
                unassigned_lists.pop(0)  # pop only after iterating over
        return article



from configparser import ConfigParser
from pathlib import Path

from ..ns_util import ns, ns_path
from .emitters import Emitter

__all__ = ["standup"]


def standup(domains_list=None, dry_config=False, verbose=True):
    ini_fn = "emitters.ini"
    ini_dir = Path(__file__).parent
    c = ConfigParser()
    with open(ini_dir / ini_fn, "r") as f:
        c.read_file(f, ini_fn)
    emitters = {}
    if domains_list is None:
        domains_list = [*ns]
    for domain in c.sections():
        if domain not in domains_list:
            continue  # skip 'DEFAULT' section
        sect = c[domain]
        ns_p = ns_path / domain
        assert ns_p.exists() and ns_p.is_dir(), f"{ns_p} not found"
        for name, emit_type in sect.items():
            if verbose:
                print(f"{name}-{emit_type}")
            directory = ns_p / name
            e = getattr(Emitter, emit_type).value
            if dry_config:
                emitter = {
                    "emitter": e,
                    "directory": directory,
                    "name": name,
                    "verbose": verbose,
                }
            else:
                emitter = e(directory=directory, name=name, verbose=verbose)
            # emitters.setdefault(domain, [])
            # emitters[domain].append(emitter)
            emitters.update({domain: emitter})
    return emitters



import logging
from datetime import datetime as dt
from functools import partial
from pathlib import Path

import dateparser
import frontmatter
import markdown
from staticjinja import Site

from ..ns_util import ns, ns_path

__all__ = ["standup"]

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

pymd_extensions = (
    "fenced_code codehilite sane_lists def_list admonition toc markdown_katex".split()
)
extension_configs = {
    "markdown_katex": {"no_inline_svg": True, "insert_fonts_css": False},
}
markdowner = markdown.Markdown(
    output_format="html5",
    extensions=pymd_extensions,
    extension_configs=extension_configs,
)

OUT_DIRNAME = "site"
TEMPLATE_DIRNAME = "src"
POST_DIRNAME = "posts"


def standup(domains_list=None, verbose=True):
    domains_list = domains_list or [*ns]
    for domain in domains_list:
        ns_p = ns_path / domain
        if not ns_p.exists() and ns_p.is_dir():
            raise ValueError(f"{ns_p} not found")
        template_dir = ns_p / TEMPLATE_DIRNAME
        if template_dir.exists():
            site_dir = ns_p / OUT_DIRNAME
            post_dir = template_dir / POST_DIRNAME
            extra_ctxs = []
            if post_dir.exists():
                post_dir_sans_drafts = [
                    a for a in post_dir.iterdir() if a.name != "drafts"
                ]
                for a in post_dir_sans_drafts:
                    logger.info(f"POST: {a}")
                post_leaf_dir = Path(POST_DIRNAME)
                post_ctxs = [
                    (str(post_leaf_dir / a.name), article) for a in post_dir_sans_drafts
                ]
                extra_ctxs.extend(post_ctxs)
                extra_ctxs.extend(
                    [
                        ("index.html", partial(indexed_articles, dir_path=post_dir)),
                        # ("index.html", partial(indexed_article_series, dir_path=post_dir)),
                    ]
                )
            cut_templates(
                template_dir=template_dir, out_dir=site_dir, contexts=extra_ctxs
            )
            logger.info(f"Built {template_dir}")


def fmt_mtime(path_to_file):
    mtime = path_to_file.stat().st_mtime
    date = dt.fromtimestamp(mtime)
    return date.strftime("%Y-%m-%dT%H:%M:%SZ")


def base(template):
    logger.info(f"Rendering {template} (base)")
    template_path = Path(template.filename)
    return {
        "template_date": fmt_mtime(template_path),
    }


def index(template):
    logger.info(f"Rendering {template} (index)")
    return {}


def indexed_article_series(template, dir_path, drop_hidden=True):
    "Sort article series by date"
    series_dict = {
        "series": sorted(
            [
                article_series(a, is_path=True)
                for a in dir_path.iterdir()
                if a.is_dir()  # sub-directory
                if (a / "index.md").exists()
                if not a.name.startswith("_")  # not partial template
            ],
            key=lambda d: dateparser.parse(d["date"]),
            reverse=True,
        )
    }
    if drop_hidden:
        series_dict = {
            "series": [
                article_series
                for article_series in series_dict["series"]
                if (
                    "hidden" not in article_series
                    or article_series["hidden"] is not True
                )
            ]
        }
    return series_dict


def indexed_articles(template, dir_path, with_series=True):
    "Sort articles by date"
    articles_dict = {
        "articles": sorted(
            [
                article(a, is_path=True)
                for a in dir_path.iterdir()
                if not a.is_dir()  # not sub-directory
                if not a.name.startswith("_")  # not partial template
            ],
            key=lambda d: dateparser.parse(d["date"]),
            reverse=True,
        )
    }
    if with_series:
        series_dict = indexed_article_series(template, dir_path)
        articles_dict = {
            "articles": sorted(
                [
                    *articles_dict["articles"],
                    *series_dict["series"],
                ],
                key=lambda d: dateparser.parse(d["date"]),
                reverse=True,
            )
        }
    return articles_dict


def article_series(template, is_path=False):
    logger.info(f"Rendering {template} (article series)")
    template_path = template if is_path else Path(template.filename)
    template_index_path = template_path / "index.md"
    if not template_index_path.exists():
        raise ValueError(
            "Metadata is not supported for non-markdown articles (index.md not found)"
        )
    md_content = frontmatter.load(template_index_path)
    metadata = md_content.metadata
    required_keys = {"title", "desc", "date"}
    if not all(k in metadata for k in required_keys):
        raise ValueError(
            f"{template=} missing one or more of {required_keys=} in index.md"
        )
    return {"url": template_path.stem, "mtime": fmt_mtime(template_path), **metadata}


def article(template, is_path=False):
    logger.info(f"Rendering {template} (article)")
    template_path = template if is_path else Path(template.filename)
    if template_path.suffix != ".md":
        raise ValueError("Metadata is not supported for non-markdown articles")
    md_content = frontmatter.load(template_path)
    # Introduces second pass on every markdown (to check for KaTeX markup)
    html_content = convert_markdown(md_content.content)
    has_katex = """<span class="katex">""" in html_content
    metadata = {**md_content.metadata, "katex": has_katex}
    required_keys = {"title", "desc", "date"}
    if not all(k in metadata for k in required_keys):
        raise ValueError(f"{template=} missing one or more of {required_keys=}")
    return {"url": template_path.stem, "mtime": fmt_mtime(template_path), **metadata}


def convert_markdown(markdown: str):
    return markdowner.convert(markdown)


def md_context(template):
    logger.info(f"Rendering {template} (md context)")
    md_content = frontmatter.load(template.filename)
    return {"post_content_html": convert_markdown(md_content.content)}


def render_md(site, template, **kwargs):
    if "/drafts/" in template.name:
        return
    logger.info(f"Rendering {template} (md)")
    # i.e. posts/post1.md -> build/posts/post1.html
    template_out_as = Path(template.name).relative_to(Path(POST_DIRNAME))
    out_parts = list(template_out_as.parts)
    is_index = template_out_as.stem == "index"
    is_multipart = len(out_parts) > 1
    if is_multipart:
        if is_index:
            out_parts.pop()  # Remove the index, giving it the parent dir as its path
        template_out_as = Path("-".join(out_parts))
    # URLs do not have to end with .html so do not add suffix other than for index file
    # Multipart indexes drop the index name but for suffix we still treat the same
    out_suffix = ".html" if (is_index and not is_multipart) else ""
    out = site.outpath / template_out_as.with_suffix(out_suffix)
    # Compile and stream the result
    site.get_template("layouts/_post.html").stream(**kwargs).dump(
        str(out), encoding="utf-8"
    )


def cut_templates(template_dir, out_dir, contexts=None, mergecontexts=True):
    default_ctxs = [
        (r".*\.(html|md)$", base),
        ("index.html", index),
        (r".*\.md", md_context),
    ]
    custom_contexts = contexts or []
    contexts = default_ctxs + custom_contexts
    site = Site.make_site(
        contexts=contexts,
        mergecontexts=mergecontexts,
        searchpath=template_dir,
        outpath=out_dir,
        rules=[(r".*\.md", render_md)],
    )
    try:
        site.render()
    except:
        logging.info("Failed to render site", exc_info=True)



from .ns_util import *  # isort: skip
from .cut import *
from .git import *
from .man import *
from .wire import *
from .yaml_util import *

__all__ = [
    "ns",
    "ns_path",
    "clone",
    "source_manifest",
    "remote_push_manifest",
    "remote_pull_manifest",
    "stash_transfer_site_manifest",
    "yaml_manifests",
    "get_yaml_manifest",
    "change_build_dir",
]



from bs4 import Tag

__all__ = ["node_html_tag", "HtmlBlock"]


def node_html_tag(node, line_no=None):
    "Should this be in scan⠶lever"
    pre_name = node.prefix._name_
    pre = node.prefix.str if node.prefix else ""
    con = node.contents if node.contents else ""
    suf = node.suffix.str if node.suffix else ""
    inner = con.strip()
    if suf != "":
        suf_span = Tag(name="span")
        suf_span.append(suf)
    else:
        suf_span = ""
    if pre_name == "SectBreak":
        t = Tag(name="hr")
        # never has inner content
    else:
        # anything with inner content goes here
        if pre_name.startswith("Header"):
            # headers can't go in inline span elements
            html_tag_name = pre_name.replace("Header", "h")
            t = Tag(name=html_tag_name)
        else:
            t = Tag(name="span")
        t.append(inner)
        t.append(suf_span)
    # Then regardless of whether/not the node has inner content:
    cls_list = ["Node", f"pre_{pre_name}"]
    if node.suffix:
        cls_list.append(f"suf_{node.suffix._name_}")
    attrs = {"class": " ".join(cls_list), "prefix": pre, "suffix": suf}
    t.attrs.update(attrs)
    if line_no is not None:
        t.attrs.update({"id": f"L{line_no}"})
    return t


class HtmlBlock:
    def __init__(self, block, lists, qa_pair=False, nested_qa_summary=False):
        self.number = block.number
        self.start_line = block.start_line
        self.end_line = block.end_line
        self.nodes = block.nodes
        self.lists = lists
        self.qa_pair = qa_pair
        self.nested_qa_summary = nested_qa_summary

    def as_soup(self):
        line_range_dict = {"from": f"{self.start_line}", "to": f"{self.end_line}"}
        attrs = {"class": "NodeBlock", "id": f"nb{self.number}", **line_range_dict}
        block_div = Tag(name="div", attrs=attrs)
        # TODO do this properly after rewriting list parsing so end lines are accurate
        # list_starts = [l.all_nodes[0].start_line for l in self.lists]
        # list_ends = [l.all_nodes[-1].end_line for l in self.lists]
        if self.qa_pair:
            self.qa_idx_pairs = {}
        for i, n in enumerate(self.nodes):
            if n.prefix is None:
                raise ValueError(f"Missing prefix for '{n.content}' (tag won't parse)")
            node_line_no = self.start_line + i
            html_node = node_html_tag(n, node_line_no)
            if self.qa_pair:
                if hasattr(n, "paired_to"):
                    paired_block_no, paired_line_no = n.paired_to
                    html_node.attrs.update({"paired_id": f"L{paired_line_no}"})
                    if self.nested_qa_summary:
                        if paired_line_no > node_line_no:
                            # Build the details tag, summary first, but don't finish it
                            html_node.name = "summary"
                            accumulated_det = Tag(name="details")
                            accumulated_det.append(html_node)
                            continue
                        else:
                            # The details can stay as a span tag and finish the summary
                            accumulated_det.append(html_node)
                            # Overwrite html_node variable to append entire summary tag
                            html_node = accumulated_det
            block_div.append(html_node)
        return block_div



from pathlib import Path
from sys import stderr

from yaml import Loader, load

from ..manifest.man import ssm
from .ns_util import ns, ns_path
from .site_yaml import SiteCI

__all__ = ["yaml2dict", "yaml_manifests", "get_yaml_manifest", "change_build_dir"]

YAML_FNAME = ".gitlab-ci.yml"


def yaml2dict(yaml_path):
    with open(yaml_path, "r") as f:
        return load(f, Loader=Loader)


def yaml_manifests(as_dicts=False, yaml_filename=YAML_FNAME):
    """
    Clone repos as per the manifest (`qu.ssm`)
    """
    ssm.check_manifest()
    df = ssm.repos_df.loc[:, ("domain", "local")]
    manifest_dict = {}
    for domain, local in df.values:
        if local:
            local_path = ns_path / domain
            yaml_path = local_path / yaml_filename
            d = yaml2dict(yaml_path)
            manifest_dict.update({domain: d})
        else:
            pass  # do not touch if not local (for now)
    if as_dicts:
        return manifest_dict
    else:
        layout_dict = {}
        for d, y in manifest_dict.items():
            try:
                layout_dict.update({d: SiteCI(y)})
            except:
                print(
                    f"Error encountered processing domain {d} (got: {layout_dict=})",
                    file=stderr,
                )
                raise
        return layout_dict


def get_yaml_manifest(domain, as_dicts=False, yaml_filename=YAML_FNAME):
    all_manifests = yaml_manifests(as_dicts=as_dicts, yaml_filename=yaml_filename)
    assert domain in all_manifests, f"No YAML manifest for '{domain=}'"
    return all_manifests.get(domain)


def change_build_dir(domain, new_name, mv=True, yaml_filename=YAML_FNAME):
    """
    For a given domain, change the directory its 'script' entry is configured
    to build from (i.e. the `subdirectory` variable in the
    `fold`⠶`site_yaml.Script.cfg` dict).

    If `mv` is `True`, the files in the current build directory will be moved
    into this new build directory. N.B. any "hidden" files whose filenames
    begin with `.` will not be moved if the current build directory is the
    top-level directory, as this would include for instance `.git/` and
    `.gitlab-ci.yml` which must be in the top-level directory to function.
    For this reason it is advised to always use a non-top-level directory
    as the build directory (to avoid having to figure out how to exclude some
    files from a directory move in the event of changing it at a later stage).
    """
    # move the files from the current build dir to the new build dir, either:
    # - simply rename the current build dir to `new_name`
    # - if no current build dir: `mkdir new_name` and `mv ./* {new_name}/`
    assert type(domain) is type(new_name) is str, f"'{domain}'/'{new_name}' aren't str"
    local_path = ns_path / domain
    yaml_path = local_path / yaml_filename
    assert yaml_path.exists(), f"No YAML file at '{yaml_path!=s}'"
    site_ci = get_yaml_manifest(domain, as_dicts=False, yaml_filename=yaml_filename)
    # If block to ensure any path error while moving files vetoes YAML modification
    if mv:
        current_build_dir = site_ci.pages.script.cfg.get(
            "subdirectory"
        )  # None if unset
        if current_build_dir:
            current_build_path = local_path / current_build_dir
            yaml_expect_msg = f"'{yaml_path=!s}' says {current_build_path=!s} exists"
            assert current_build_path.exists(), f"{yaml_expect_msg}, but it doesn't!"
            # otherwise `current_build_path` exists so just rename the directory
            current_bp_par = current_build_path.parent
            new_build_path = current_build_path.rename(current_bp_par / new_name)
            print(
                f"Moved build path for {domain} from '{current_build_dir}' "
                f"--> {new_build_path}",
                file=stderr,
            )
        else:
            # implicitly: `current_build_path` = `local_path`
            new_build_path = local_path / new_name
            assert not new_build_path.exists(), f"{new_build_path=!s} already exists!"
            new_build_path.mkdir()
            # do not move "hidden" files (those whose filenames begin with a period)
            movable_files = [
                x for x in local_path.iterdir() if not x.name.startswith(".")
            ]
            for source in movable_files:
                if source.is_dir() and source.name == new_name:
                    continue  # cannot move the new directory into itself!
                dest = new_build_path / source.name
                source.replace(dest)
            print(f"Created build path for {domain} at {new_build_path}", file=stderr)
    # change build directory name in the yaml's script
    site_ci.pages.script.cfg.update({"subdirectory": Path(new_name)})
    # write the SiteCI as YAML to `.gitlab-ci.yml`
    with open(yaml_path, "w") as f:
        f.write(site_ci.as_yaml)



from pandas import DataFrame

from ...fold.ns_util import ns
from ...manifest.parsing import read_man, read_man_df
from .blockelems import *  # temporary
from .docelems import DocLists
from .git import _active_branch, _has_clean_wt
from .lists import BlockList, SepBlockList, parse_nodes_to_list
from .structure import BlockDoc

__all__ = ["Doc"]


class PartsList(list):
    """
    Simple list class which provides a pandas DataFrame method which will be
    bound to the `Doc` instance (via a descriptor).
    """

    def __init__(self, parts, part_keys=None):
        self.extend(parts)
        self.part_keys = part_keys

    def as_df(self):
        datadict = {k: [n[i] for n in self] for (i, k) in enumerate(self.part_keys)}
        return DataFrame.from_dict(datadict)


class Doc(BlockDoc):
    def __init__(self, lines, listparseconfig=None):
        super().__init__(lines)  # block tokenisation pass, creating nodes property
        if listparseconfig is None:
            listparseconfig = {"listclass": BlockList}
        elif listparseconfig.get("listclass") == "auto":
            listparseconfig.update({"listclass": BlockList})
        elif "sep" in listparseconfig and "listclass" not in listparseconfig:
            # helper: do not require passing the list class itself, assume it from `sep`
            sep_listconfig_keys = ["sep", "headersep", "labels"]
            cfg = {
                k: v for (k, v) in listparseconfig.items() if k in sep_listconfig_keys
            }
            for k in sep_listconfig_keys:
                if k in listparseconfig:
                    del listparseconfig[k]
            listparseconfig.update({"listclass": SepBlockList, "listconfig": cfg})
        self._parse(listparseconfig)  # tokenised block parsing, creating lists property
        if hasattr(self, "all_parts") and self.all_parts.part_keys:
            self.as_df = self.all_parts.as_df
        if self.list:
            self.repos = property(read_man).__get__(self)
            self.repos_df = property(read_man_df).__get__(self)

    def _parse(self, listparseconfig=None):
        "May add other configs later, but for now just wrap the lists method."
        # populate the `lists` property by parsing all blocks' nodes
        self._parse_lists(**listparseconfig)  # expand out dict as named arguments

    def _parse_lists(
        self, listclass=None, part_keys=None, listconfig=None, strict_list_breaks=True
    ):
        """
        Note that for separator-delimited lists, `labels` (in `listconfig`) is for
        setting the names of node attributes of each delimited value, while
        `part_keys` sets column names on the pandas DataFrame.
        """
        # TODO strict_list_breaks param
        all_blocklists = []
        for block in self.blocks:
            nodes = block.nodes
            # yields blocklist objects
            bl_generator = parse_nodes_to_list(nodes, listconfig, listclass)
            blocklists = list(bl_generator)  # exhaust generator
            all_blocklists.extend(blocklists)  # flat list: all BlockList objects in Doc
        # `DocLists` object from list of BlockList objects
        self.lists = DocLists(all_blocklists)
        if listclass is SepBlockList:
            pk = part_keys
            # N.B. maybe use an Enum rather than have to pass actual class?
            self.all_parts = PartsList(
                [
                    n.parts
                    for l in self.lists
                    for n in (l.all_nodes if l.has_sep_header else l.nodes)
                ],
                part_keys=pk,
            )
        # TODO: wasteful to store nodes twice: just store index for header/list items

    @property
    def lists(self):
        # What I really want here is the union, over all blocks in the Doc,
        # of each's block's lines' output from parse_nodes_to_list
        # which will probably necessitate a more complicated structure,
        # in some dedicated class which manages lists by block
        return self._lists

    @property
    def list(self):
        if len(self.lists) == 1:
            return self.lists[0]
        else:
            return None

    @lists.setter
    def lists(self, ll):
        self._lists = ll

    @property
    def _doc_repr(self):
        "Private property to enable access from MMD superclass"
        block_repr = super().__repr__()
        cont = ", containing "
        parsed_list_repr = self.lists.repr_str if self.lists else ""
        parsed_repr = f"{cont}{parsed_list_repr}" if self.lists else ""
        return f"{block_repr}{parsed_repr}"

    def __repr__(self):
        return self._doc_repr

    def check_manifest(self, add_before_check=True):
        domains = self.repos_df.domain
        self.repos_df["branch"] = [_active_branch(d) for d in domains]
        self.repos_df["local"] = [domain in ns for domain in domains]
        self.repos_df["clean"] = [
            _has_clean_wt(d, add_before_check=add_before_check)
            for d in domains
            if d in ns
        ]



from pandas import concat, merge

from ..fold.ns_util import ns_path
from ..manifest.man import ssm
from ..manifest.namings import alias_df

__all__ = ["write_man_README"]


def process_df_row(r, remote_template="https://gitlab.com/{}/{}.gitlab.io"):
    namespaced = f"{r.namespace} ⠶ {r.alias if r.alias else r.domain}"
    remote_url = remote_template.format(*[r.repo_name] * 2)
    return f"- `{r.domain}`: [{namespaced}]({remote_url})"


def write_man_README(title="spin.systems"):
    """
    Store a representation with links from the manifest `qu.ssm`
    """
    df = merge(alias_df, ssm.as_df())
    domain_vals = ["cal", "log", "conf", "pore", "ocu", "arc", "qrx", "erg", "opt"]
    domain_vals += ["poll", "arb", "reed", "noto", "plot", "doc", "labs"]
    df = concat([df.query(f"domain == '{d}'") for d in domain_vals])
    list_entries = list(map(lambda r: process_df_row(r[1]), df.iterrows()))
    ss_url = "https://gitlab.com/{}/{}.gitlab.io".format(*["spin.systems"] * 2)
    README_md_lines = [f"# {title}", "", f"spin.systems: [`ss`]({ss_url})", ""]
    README_md_lines += list_entries
    README_md = "\n".join(README_md_lines) + "\n"
    with open(ns_path / "README.md", "w") as f:
        f.write(README_md)
    return



from .tokens import Prefix, Suffix, tokenise_line

__all__ = ["BlockDoc", "NodeBlock"]


class BlockDoc:
    """
    Document formed by a list of one or more `NodeBlock` elements,
    created upon reading file lines (newlines will be stripped).
    """

    def __init__(self, lines):
        self._initblocks(lines)

    def _add_nodeblock(self, lines, line_no, block_no):
        self.blocks.append(NodeBlock(lines, line_no, block_no))

    def _initblocks(self, doc_lines):
        doc_lines = [l.rstrip("\n") for l in doc_lines]
        self.blocks = []
        current_line_block = []
        for line_no, l in enumerate(doc_lines):
            if l == "" and current_line_block:
                self._add_nodeblock(current_line_block, line_no, self.n_blocks)
                current_line_block = []
            else:
                current_line_block.append(l)
        if current_line_block:
            self._add_nodeblock(current_line_block, line_no + 1, self.n_blocks)

    @property
    def blocks(self):
        return self._blocks

    @blocks.setter
    def blocks(self, blocks):
        self._blocks = blocks

    @property
    def n_blocks(self):
        return len(self.blocks)

    def __repr__(self):
        s = "s" if (self.n_blocks != 1) else ""
        return f"Document of {self.n_blocks} block{s}"


class NodeBlock:
    def __init__(self, block_lines, line_no, block_no):
        self._nodes = []  # private property
        self.start_line = line_no - len(block_lines) + 1
        self.end_line = line_no
        self.number = block_no
        self.tokenise_lines(block_lines)

    @property
    def nodes(self):
        return self._nodes

    def add_node(self, node):
        self._nodes.append(node)

    @property
    def _nodecount(self):
        return len(self.nodes)

    def __repr__(self):
        return (
            f"Block {self.number} of {self._nodecount} "
            f"nodes (L{self.start_line}-L{self.end_line})"
        )

    def tokenise_lines(self, block_lines):
        "Populate nodes property"
        if block_lines[0].endswith("\n"):
            raise ValueError("Strip off newlines before passing into `NodeBlock`!")
        tokenised = []
        for i, l in enumerate(block_lines):
            line_no = self.start_line + i
            n = tokenise_line(l, line_no, self.number, seen=tokenised)
            tokenised.append(n)
        # Since preceding lines' nodes are modified in the processing
        # of subsequent lines, only add nodes to block after finishing.
        for node in tokenised:
            self.add_node(node)
        return



from pandas import concat

from ..fold import ns_path
from ..scan.io import mmd

__all__ = ["m", "alias_df"]

pk = ["domain", "alias"]  # `labels` to name df cols and `part_keys` for attr names
cfg = {"sep": "=", "headersep": True, "labels": pk, "part_keys": pk}
m = mmd(ns_path / "alias.mmd", listparseconfig=cfg)
# aliases = {l.header.contents: [n.contents for n in l.nodes] for l in m.lists}
alias_df = {
    (l.header.domain, l.header.alias): l.as_df(forbid_header=True) for l in m.lists
}
for k in alias_df:
    ns_long, ns_short = k
    alias_df.get(k).insert(0, "namespace", ns_short)
    alias_df.get(k).insert(0, "namespace_full", ns_long)
alias_df = concat(alias_df.values()).reset_index().drop("index", 1)
alias_df.namespace = alias_df.namespace.astype("category")
alias_df.namespace_full = alias_df.namespace_full.astype("category")



from .elems import BaseElem

__all__ = ["BlockElem", "BlockList"]


class BlockElem(BaseElem):
    "Base class for any block-level element."
    # def __init__(self, nodelist):
    #    super().__init__(nodelist)


class BlockList(BlockElem):
    """
    A block-level list would be a list of `ListInit`-prefixed
    nodes and could have a `ListInit`-suffixed header node.

    If initialised in the midst of a different element type, pass its
    element type (block- or doc-level) through `parent_elem` argument.
    """

    def __init__(self, nodes, header=None, parent_elem=None):
        super().__init__(nodes)  # once sublist of nodes determined
        self.header = header
        self.parent_elem = parent_elem

    @property
    def header(self):
        return self._header

    @header.setter
    def header(self, h):
        self._header = h

    @property
    def all_nodes(self):
        return [self.header, *self.nodes] if self.header else self.nodes

    def __repr__(self):
        headered = ("H" if self.header else "Unh") + "eadered"
        return f"{headered} list with {len(self.nodes)} items"



from itertools import starmap
from os import utime
from pathlib import Path
from shutil import rmtree
from subprocess import run
from sys import stderr

from git import Repo

from ..manifest.man import ssm
from . import cut
from .ns_util import ns, ns_path, pre_existing_ns_p

__all__ = [
    "clone",
    "source_manifest",
    "GitNews",
    "preprocess_domains_list",
    "remote_push_manifest",
    "remote_pull_manifest",
    "stash_transfer_site_manifest",
]


def clone(url, as_name, wd=ns_path, update_man=True, use_git_mtime=False):
    """
    Clone a repo from ``url`` giving it the name ``as_name``. If ``use_git_mtime`` is
    ``True``, then modify the access and modified timestamps (atime and mtime) of all
    files in the git repo's working tree to be the most recent commit in the git log.
    """
    clone_path = wd / as_name
    repo = Repo.clone_from(url, to_path=clone_path)
    ns.refresh()
    if use_git_mtime:
        for n in repo.tree().list_traverse():
            filepath = clone_path / n.path
            unixtime = repo.git.log("-1", "--format='%at'", "--", n.path).strip("'")
            if not unixtime.isnumeric():
                raise ValueError(
                    f"git log gave non-numeric timestamp {unixtime} for {n.path}"
                    f" during the cloning of {url}"
                )
            utime(filepath, times=(int(unixtime), int(unixtime)))
    if update_man:
        ssm.check_manifest()
    return


def source_manifest(use_git_mtime=True):
    """
    Clone repos as per the manifest (`qu.ssm`), checking out the last branch listed in
    the branches field [space-separated names] if distributed. If ``use_git_mtime`` is
    ``True`` (the default), then modify the mtime of the cloned files with the time of
    their last commit in the git log, rather than the one created during cloning (which
    would otherwise be the current date and time).
    """
    df = ssm.repos_df.loc[:, ("domain", "git_url", "branches")]
    for domain, url, branches in df.values:
        if (ns_path / domain).exists():
            continue  # simply do not touch for now
        try:
            clone(url, as_name=domain, update_man=False, use_git_mtime=use_git_mtime)
        except Exception as e:
            print(f"Failed on {url}: {e}", file=stderr)
    ssm.check_manifest()
    return


class GitNews:
    status_mapping = {
        "M": "Changed",
        "A": "Added",
        "D": "Deleted",
        "R": "Renamed",
        "C": "Copied",
        "U": "Updated",
    }

    def __init__(self, repo, abbreviate_at=2):
        """
        Given a repo object, generate the `git status --porcelain` output
        and parse it to give an overview of what changed.
        """
        self._news = repo.git.status("--porcelain")
        self.gnew = self._news.split("\n")
        self.status_lists = [*map(self.get_status, self.status_mapping)]
        self.abbreviate_at = abbreviate_at

    @property
    def status_labels(self):
        return [*self.status_mapping.values()]

    @property
    def all_reports(self):
        return [
            *starmap(self.report_status, zip(self.status_labels, self.status_lists))
        ]

    def __news_repr__(self):
        report = [". ".join(stats) for stats in self.all_reports if stats]
        return ". ".join(report)

    def get_status(self, letter_code):
        return [stat[3:] for stat in self.gnew if stat[0] == letter_code]

    def abbreviate_status(self, status):
        return [f"{len(status)} files"] if len(status) > self.abbreviate_at else status

    def report_status(self, action, status):
        "Returns e.g. 'Changed fileA.' or 'Changed 3 files.' as a string."
        status = self.abbreviate_status(status) if self.abbreviate_at else status
        return [f"{action} ".join(["", ", ".join(l)]) for l in [status] if status]


def preprocess_domains_list(specific_domains):
    if specific_domains is None:
        domains = ssm.repos_df.domain
    else:
        if type(specific_domains) is list:
            domains = specific_domains
        elif type(specific_domains) is str:
            domains = [specific_domains]
        else:
            raise TypeError(f"Unexpected type for {specific_domains=}")
    return domains


def remote_push_manifest(
    commit_msg=None, refspec=None, specific_domains=None, prebuild=True
):
    """
    Run `git add --all` on each repo in the manifest (`qu.ssm`),
    i.e. apex and all subdomains, then `git commit -m "..."`
    where `...` is replaced by `commit_msg` or auto-generated
    if no commit is available.
    """
    domains = preprocess_domains_list(specific_domains)
    for domain in domains:
        repo_dir = ns_path / domain
        if not repo_dir.exists():
            print(f"Skipping '{repo_dir=!s}' (doesn't exist)", file=stderr)
            continue  # simply do not touch for now
        repo = Repo(repo_dir)
        repo.git.add("--all")
        if not repo.is_dirty():
            print(f"Skipping '{repo_dir=!s}' (working tree clean)", file=stderr)
            continue  # repo has no changes to untracked files, skip it
        if prebuild:
            cut.standup(domains_list=[domain])
            repo.git.add("--all")
        repo_commit_msg = commit_msg
        if repo_commit_msg is None:
            news = GitNews(repo)
            repo_commit_msg = news.__news_repr__()
        if repo_commit_msg == "":
            msg = f"git repo stage added to at '{repo_dir=!s}'"
            raise ValueError("{msg} - aborting commit (empty commit message)")
        else:
            repo.git.commit("-m", repo_commit_msg)
            print(f"Commit [{repo_dir=!s}] ⠶ {repo_commit_msg}", file=stderr)
            origin = repo.remotes.origin
            origin.push(refspec=refspec)
            print(f"⇢ Pushing ⠶ {origin.name}", file=stderr)
    ssm.check_manifest()
    return


def remote_pull_manifest(specific_domains=None):
    """
    Run `git pull` on each repo in the manifest (`qu.ssm`),
    i.e. apex and all subdomains.

    No merge method is specified (unclear whether this will be necessary).

    See here if there are problems:
    https://stackoverflow.com/questions/36891470/how-to-pull-with-gitpython
    """
    domains = preprocess_domains_list(specific_domains)
    for domain in domains:
        print(f"Examining {domain}...", file=stderr)
        repo_dir = ns_path / domain
        if not repo_dir.exists():
            print(f"Skipping '{repo_dir=!s}' (doesn't exist)", file=stderr)
            continue  # simply do not touch for now
        repo = Repo(repo_dir)
        origin = repo.remotes.origin
        origin.pull()  # not checked if returned Pull object stores useful info
        print(f"⇢ Pulling ⠶ {origin.name}", file=stderr)
    ssm.check_manifest()
    return


def copy_static_assets(repo_dir, from_name="static", to_name="site", purge=False):
    static_dir = repo_dir / from_name
    site_dir = repo_dir / to_name
    if purge and site_dir.exists():
        rmtree(site_dir)
    site_dir.mkdir(exist_ok=not purge)
    clobber_flag = [] if purge else ["--no-clobber"]
    cp_cmd = ["cp", "-r", *clobber_flag, f"{static_dir!s}/.", f"{site_dir!s}"]
    if run(cp_cmd).returncode != 0:
        raise ValueError(f"Failed to copy {static_dir=} to {site_dir=}")


def stash_transfer_site_manifest(
    commit_msg=None,
    stash_pathspec="site/",
    checkout_branch="www",
    purge=True,
    specific_domains=None,
    reset_branch=False,
):
    """
    Stash [push] the ``stash_pathspec`` (default: "site/"), then ``git clean`` any
    changes and checkout the ``checkout_branch`` (default: "www"). Purge the stashed
    pathspec (if any files exist there) and then pop the stash to place the generated
    files there (i.e. delete anything that was there beforehand so that it's not merged,
    but completely overwritten, avoiding stale files being deployed). Lastly, add the
    stash pathspec to the git index before committing it (default ``commit_msg`` is
    ``None``: use an auto-generated commit message) and then pushing the commit to the
    ``checkout_branch``. If ``reset_branch`` is ``True`` (default: ``False``) then the
    original branch is checked out (it should be clean as the pathspec stashed from the
    original branch and pushed to the ``checkout_branch`` were the same).

    To be used after a site has been built (implicitly on the master branch).
    """
    domains = preprocess_domains_list(specific_domains)
    for domain in domains:
        repo_dir = ns_path / domain
        if not repo_dir.exists():
            print(f"Skipping '{repo_dir=!s}' (doesn't exist)", file=stderr)
            continue  # simply do not touch for now
        elif (repo_dir / "static").exists():
            copy_static_assets(repo_dir)
        repo = Repo(repo_dir)
        initial_repo_branch = repo.active_branch.name
        # Stash the desired changes (only in the given pathspec)
        stashed_msg = repo.git.stash("push", "-a", "--", stash_pathspec)
        if stashed_msg == "No local changes to save":
            print(domain, stashed_msg, "⠶ Skipping", file=stderr)
            continue
        else:
            print("Stashed changes for", domain, file=stderr)
        # Clean away any potentially undesired changes
        repo.git.clean("-fdx")
        repo.git.checkout(checkout_branch)
        if purge:
            paths_to_rm = stash_pathspec.split(" ")  # no spaces in file/dir names!
            for rm_path in paths_to_rm:
                rm_p = repo_dir / rm_path
                if rm_p.exists():
                    rmtree(rm_p) if rm_p.is_dir() else rm_p.unlink()
                else:
                    print(f"Skipping '{rm_path=!s}' (doesn't exist)", file=stderr)
        repo.git.stash("pop")
        repo.git.add(stash_pathspec)
        if not repo.is_dirty():
            print(f"Skipping '{repo_dir=!s}' (working tree clean)", file=stderr)
            continue  # repo has no changes to tracked pathspec files, skip it
        repo_commit_msg = commit_msg
        if repo_commit_msg is None:
            news = GitNews(repo)
            repo_commit_msg = news.__news_repr__()
        if repo_commit_msg == "":
            msg = f"git repo stage added to at '{repo_dir=!s}'"
            raise ValueError("{msg} - aborting commit (empty commit message)")
        else:
            repo.git.commit("-m", repo_commit_msg)
            print(f"Commit [{repo_dir=!s}] ⠶ {repo_commit_msg}", file=stderr)
            origin = repo.remotes.origin
            origin.push(refspec=checkout_branch)
            print(f"⇢ Pushing ⠶ {origin.name} ({checkout_branch})", file=stderr)
        if reset_branch:
            repo.git.checkout(initial_repo_branch)
    ssm.check_manifest()
    return



from bs4 import Tag

__all__ = ["CustomHtmlTag", "Attrs", "PartialAttrs"]

######## General utility classes for tags and attrs


class CustomHtmlTag(Tag):
    "Base class defining as_str for a bs4.Tag (may be overridden)"

    def as_str(self):
        return str(self)  # .prettify()


class Attrs(dict):
    "For passing into bs4.Tag constructor"

    def __init__(self, attr_dict):
        self.update({"attrs": attr_dict})


class PartialAttrs(Attrs):
    "Store a dict as an object property and pass a dict to create an attrs dict"

    def __init__(self, custom_attrs):
        super().__init__({**self.default_attrs, **custom_attrs})

    @property
    def default_attrs(self):
        return self._default_attrs if hasattr(self, "_default_attrs") else dict()

    @default_attrs.setter
    def default_attrs(self, default_dict):
        self._default_attrs = default_dict



from collections.abc import Sequence
from pathlib import Path

from bs4 import Tag

from .html_elem_util import BreadCrumb, NavHeader, NavLinkList
from .html_page_util import HtmlPage
from .html_util import Attrs, PartialAttrs

__all__ = ["IndexNav", "IndexUl", "EmittedIndex", "IntermedDirIndex", "WireIndex"]


class BaseIndexPage(HtmlPage):
    def __init__(self, nav, head_params={}):
        head_params["sakura"] = False
        self._nav_tag = nav
        super().__init__(self._index_content, self.rel_depth, head_params)

    @property
    def _default_css_param(self):
        "Overrides HtmlPage property method of the same name, setting head styles"
        return ["styles", "index"]

    @property
    def rel_depth(self):
        return len(self.rel_dir.parts)

    @property
    def head_params(self):
        return self._head_params

    @head_params.setter
    def head_params(self, params):
        self._head_params = params

    @property
    def _index_content(self):
        nh_sec = Tag(name="section", attrs={"id": "nh"})
        up_a = Tag(name="a", attrs={"href": "..", "id": "up_arr"})
        nh_sec.append(up_a)
        if self.nav_header:
            nh_sec.append(self.nav_header)
        return [nh_sec, self._nav_tag]

    @property
    def nav_header(self):
        if hasattr(self, "_nh"):
            return self._nh
        else:
            return None

    @nav_header.setter
    def nav_header(self, h):
        self._nh = h

    @property
    def nav_header_attrs(self):
        if hasattr(self, "_nh_attrs"):
            return self._nh_attrs
        else:
            return dict()

    @nav_header_attrs.setter
    def nav_header_attrs(self, attrs):
        self._nh_attrs = attrs

    @property
    def breadcrumb_attrs(self):
        if hasattr(self, "_bc_attrs"):
            return self._bc_attrs
        else:
            return dict()

    @breadcrumb_attrs.setter
    def breadcrumb_attrs(self, attrs):
        self._bc_attrs = attrs


class IndexNav(PartialAttrs):
    "Partially set Attrs wrapper class for index page nav elements"

    def __init__(self, attrs):
        self.default_attrs = {"class": "index"}
        super().__init__(attrs)


class IndexUl(PartialAttrs):
    "Gratuitous wrapper class (currently doing nothing but may in future)"

    def __init__(self, attrs):
        # self.default_attrs = {}
        super().__init__(attrs)


class BaseWireIndexPage(BaseIndexPage):
    "Wrapper class to set the nav header text for breadcrumbs"

    def __init__(self, *args):
        self.nav_header_attrs = Attrs(
            {"id": "nav_header", "depth": f"{self.rel_depth}"}
        )
        self.create_header()
        super().__init__(*args)

    def create_header(self):
        if self.nav_header is not None:
            return  # forbid overwriting nav header by repeated call to this function
        nh_size = 3
        default_bc_attrs = {"crumb_sep": "⠶", "id": "crumbs"}
        bc_attrs = {**default_bc_attrs, **self.breadcrumb_attrs}
        crumb_sep = bc_attrs.pop("crumb_sep")
        bc = Tag(name="ul", attrs=bc_attrs)
        breadcrumbs = []
        for i, x in enumerate(self.rel_dir.parts):
            if i == len(self.rel_dir.parts) - 1:
                # final tuple: omit link off last crumb (as it is to current page)
                bc_tuple = (x,)
            else:
                bc_tuple = (x, Path(*[".."] * (len(self.rel_dir.parts) - i)) / x)
            breadcrumbs.append(bc_tuple)
        for i, c in enumerate(breadcrumbs):
            t = Tag(name="li")
            if i > 0:
                t.attrs.update({"separator": crumb_sep})
            t.append(BreadCrumb(*c))
            bc.append(t)
        self.nav_header = NavHeader(nh_size, bc, self.nav_header_attrs)


class EmittedIndex(BaseWireIndexPage):
    def __init__(self, files, rel_dir, head_params={}):
        self.rel_dir = Path("wire") / rel_dir
        self.head_params = head_params
        nav_params = IndexNav({"id": "file_index"})
        ul_params = IndexUl({"id": "files"})
        filenames = list(map(self.process_filename, files))
        nav = NavLinkList(filenames, files, nav_params, ul_params)
        super().__init__(nav)

    @staticmethod
    def process_filename(f):
        fstr = str(f)
        if any(fstr.startswith(f"{d}") for d in range(10)):
            pass
        return fstr.replace("_", " ").replace("-", " ")


class IntermedDirIndex(BaseWireIndexPage):
    def __init__(self, subdirs, rel_dir, head_params={}):
        self.rel_dir = Path("wire") / rel_dir
        self.head_params = head_params
        nav_params = IndexNav({"id": "mid_index"})
        ul_params = IndexUl({"id": "index", "class": f"lvl_{self.rel_depth}"})
        dirnames = list(map(self.process_dirname, subdirs))
        nav = NavLinkList(dirnames, subdirs, nav_params, ul_params)
        # self.nav_header_attrs = Attrs({"class": "breadcrumbs bc_intermed"})
        super().__init__(nav)
        # self.nav_header.string = " ⠶ dir_beep_boop" # breadcrumbs go here
        # self.nav_header.append(" ...beep boop?")

    @staticmethod
    def process_dirname(d):
        return str(d)


class WireIndex(BaseWireIndexPage):
    def __init__(self, subdirs, rel_dir, head_params={}):
        self.rel_dir = Path("wire") / rel_dir
        self.head_params = head_params
        nav_params = IndexNav({"id": "wire_index"})
        ul_params = Attrs({"id": "wires"})
        link_params = Attrs({"class": "wirelink"})
        dirnames = list(map(self.process_dirname, subdirs))
        nav = NavLinkList(dirnames, subdirs, nav_params, ul_params, link_params)
        super().__init__(nav)
        # self.nav_header.append(" ...beep boop?")

    @staticmethod
    def process_dirname(d):
        "Prefix the year shortname with 20 to indicate the full year"
        return f"20{d}"



from enum import IntEnum
from pathlib import Path

from . import __path__ as _dir_nspath

__all__ = ["ql_path"]
ql_path = Path(list(_dir_nspath)[0])


class classproperty(property):
    "used for prefix dicts"

    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


class MonthEnum(IntEnum):
    jan = 1
    feb = 2
    mar = 3
    apr = 4
    may = 5
    jun = 6
    jul = 7
    aug = 8
    sep = 9
    oct = 10
    nov = 11
    dec = 12


num2month_dict = dict([(x, y.name) for x, y in MonthEnum._value2member_map_.items()])

num2alphanum_monthdict = {k: f"{k:02}{v}" for k, v in num2month_dict.items()}
alphanum2num_monthdict = {v: k for k, v in num2alphanum_monthdict.items()}



from enum import Enum, IntEnum

from ...fold import ns_path
from ...manifest.namings import alias_df
from .routing import routing_df

__all__ = ["AddressPath"]


class AddressPart(str):
    def __new__(cls, *args, **kw):
        return super().__new__(cls, *args, **kw)


class NameSpaceString(AddressPart):
    pass


class DomainString(AddressPart):
    pass


class DomainAliasString(AddressPart):
    pass


class DigitStringPart(AddressPart):
    @property
    def int(self):
        return self._int

    @int.setter
    def int(self, i):
        assert isinstance(i, int), f"Not an integer: '{type(i)}'"
        self._int = i


class YyDigitString(DigitStringPart):
    def __init__(self, y):
        assert len(y) == 2, f"Expected a 2-digit string, got '{y}'"
        self.int = int(y)  # expect e.g. "20" to indicate the year 2020

    @property
    def year(self):
        return 2000 + self.int


class MonthInt(IntEnum):
    jan = 1
    feb = 2
    mar = 3
    apr = 4
    may = 5
    jun = 6
    jul = 7
    aug = 8
    sep = 9
    oct = 10
    nov = 11
    dec = 12


class MonthUtilMixin:
    @staticmethod
    def month_int_to_enum(month_int):
        return MonthInt._value2member_map_.get(month_int)

    @staticmethod
    def month_str_to_enum(month_str):
        return MonthInt._member_map_.get(month_str)

    @staticmethod
    def validate_month_int(month_int):
        assert 1 <= month_int <= 12, f"Expected integer from 1-12, got '{month_int}'"
        return


class MmDigitString(DigitStringPart, MonthUtilMixin):
    def __init__(self, m):
        mi = int(m)
        self.validate_month_int(mi)
        self.int = mi  # expect e.g. "10" to indicate the month November

    @property
    def as_enum(self):
        return self.month_int_to_enum(self.int)

    @property
    def as_str(self):
        return self.as_enum.name


class mmmString(AddressPart, MonthUtilMixin):
    def __init__(self, m):
        mi = self.month_str_to_enum(m).value
        assert 1 <= mi <= 12, f"Expected an integer from 1 to 12, got '{mi}'"
        self.int = mi  # expect e.g. "10" to indicate the month November

    @property
    def as_enum(self):
        return self.month_int_to_str(self.int)

    @property
    def as_str(self):
        return self


class DdDigitString(DigitStringPart):
    def __init__(self, d):
        di = int(d)
        self.validate_day_int(di)  # weakly validate (presume calendar obeyed)
        self.int = di  # expect e.g. "1" to indicate the 1st day of the month

    @staticmethod
    def validate_day_int(day_int):
        assert 1 <= day_int <= 31, f"Expected day number from 1-31, got {day_int}"


# TODO: alternative DdDigitString class, for {01,02,03} rather than {1,2,3}


class FileIntString(DigitStringPart):
    def __init__(self, f):
        self.int = int(f)


class AddressParts(Enum):
    NameSpace = NameSpaceString
    Domain = DomainString
    DomainAlias = DomainAliasString
    YY = YyDigitString
    MM = MmDigitString
    mmm = mmmString
    DD = DdDigitString
    FileInt = FileIntString


def pop_part(part_list):
    try:
        return part_list.pop(0)
    except IndexError as e:
        return e


def interpret_filepath(address_path):
    if type(address_path) is str:
        address_path = AddressPath(address_path)
    namespace = address_path.namespace
    domain = address_path.domain
    route_query = f"namespace == '{namespace}' and domain == '{domain}'"
    # route_result = routing_df.query(route_query)
    custom_route = routing_df.query(route_query).route.item()
    route = custom_route if custom_route else domain
    domain_filepath = (ns_path / route).resolve()
    has_y, has_m, has_d, has_f = [
        hasattr(address_path, attrib) for attrib in "year month day fileint".split()
    ]
    # Delayed to avoid circular import
    from ...fold.wire import standup

    standup_config = standup(domains_list=[domain], dry_config=True, verbose=False)
    if domain in standup_config:
        # The domain is being served, so get the directory from which it is served
        domain_filepath = standup_config.get(domain).get("directory")
    if has_y:
        yeardir = domain_filepath / address_path.year
        if has_m:
            month_str = f"{address_path.month.int:02d}{address_path.month.as_str}"
            monthdir = yeardir / month_str
            if has_d:
                daydir = monthdir / address_path.day
                if has_f:
                    if daydir.is_dir():
                        # Will raise StopIteration exception if glob comes back empty
                        p = next(daydir.glob(f"{address_path.fileint}*"))
                    else:
                        raise FileNotFoundError(f"Expected a directory at {daydir}")
                else:
                    p = daydir
            else:
                p = monthdir
        else:
            p = yeardir
    else:
        p = domain_filepath
    return p


class AddressPath(list):
    def __init__(self, address_str, sep="⠶"):
        self.sep = sep
        self.parts = SeparatedString(address_str, sep=sep)

    @property
    def as_str(self, sep=None):
        if sep is None:
            sep = self.sep
        return sep.join(self)

    @property
    def filepath(self):
        return interpret_filepath(self.as_str)

    @property
    def parts(self):
        return self._parts

    @parts.setter
    def parts(self, p):
        """
        Setting a list of `parts` triggers parsing those part strings into the typed
        `AddressParts` and appending them as the main `AddressPath` list items. The
        parts are stored but not appended to the path until validated.
        """
        self._parts = p
        self._parse()

    def _parse(self, strict=True):
        """
        If parsing strictly, expect either:
          - ∫⠶log
          - ∫⠶log⠶20⠶10⠶25
          - ∫⠶log⠶20⠶10⠶25⠶0
        Else if not `strict`, just parse as much as possible after the first part.
        """
        pp = self.parts.copy()  # copy so popping can indicate if reached final part
        if strict:
            assert len(pp) in (2, 5, 6), f"Bad address length: {len(pp)}"
        ns = pp.pop(0)  # Permit IndexError here if parts is an empty list
        assert ns in alias_df.namespace.cat.categories, f"Bad namespace: '{ns}'"
        self.namespace = AddressParts.NameSpace.value(ns)
        self.append(self.namespace)  # append typed string to path
        ns_query = f"namespace == '{self.namespace}'"
        if isinstance(d := pop_part(pp), IndexError):
            if strict:
                raise d  # we strictly want a domain in the address
            else:
                return  # reached end of the parts list so finish parsing without error
        elif d in (ns_df := alias_df.query(ns_query)).alias.to_list():
            alias = AddressParts.DomainAlias.value(d)
            # note the following will raise ValueError if multiple dealiased values
            [dealiased] = ns_df[ns_df.alias == d].domain.to_list()
            domain = AddressParts.Domain.value(dealiased)
            self.domain, self.domain_alias = dealiased, alias
            self.append(self.domain_alias)
        elif d in ns_df.domain.to_list():
            domain = AddressParts.Domain.value(d)
            # if an alias exists but isn't being used, just ignore it here (as `None`)
            self.domain, self.domain_alias = domain, None
            self.append(self.domain)
        else:
            raise ValueError("Couldn't parse address {namespace}⠶{domain|alias}")
        if isinstance(y := pop_part(pp), IndexError):
            # 2 is a permitted number of address parts, so no need to check `strict`
            return
        else:
            self.year = AddressParts.YY.value(y)
            self.append(self.year)
        if isinstance(month := pop_part(pp), IndexError):
            if strict:
                raise month  # we strictly want a full date if we saw a year
            else:
                return  # reached end of the parts list so finish parsing without error
        else:
            if month.isnumeric():
                self.month = AddressParts.MM.value(month)
            else:
                self.month = AddressParts.mmm.value(month)
            self.append(self.month)
        if isinstance(day := pop_part(pp), IndexError):
            if strict:
                raise day  # we strictly want a full date if we saw a year and month
            else:
                return  # reached end of the parts list so finish parsing without error
        else:
            self.day = AddressParts.DD.value(day)
            self.append(self.day)
        if isinstance(f := pop_part(pp), IndexError):
            # 5 is a permitted number of address parts, so no need to check `strict`
            return
        else:
            self.fileint = AddressParts.FileInt.value(f)
            self.append(self.fileint)
        if strict:
            assert len(pp) == 0, f"Strict parsing enabled but extra parts in path: {pp}"
        else:
            self.extend(pp)  # just... add extra to path untyped? ¯\_(ツ)_/¯
        return

    @classmethod
    def from_parts(cls, domain, ymd=None, n=None, sep="⠶"):
        ns = alias_df[alias_df.domain.eq(domain)].namespace.item()
        path_parts = [ns, domain]
        if ymd:
            y, m, d = map(str, ymd)
            y = y[-2:]
            m = m.zfill(2)
            d = d.zfill(2)
            path_parts.extend([y, m, d])
        if n:
            path_parts.append(str(n))
        return cls(sep.join(path_parts))


class SeparatedString(list):
    def __init__(self, sep_str, sep):
        self.extend(sep_str.split(sep))



from pandas import DataFrame

from .blockelems import BlockList
from .tokens import Prefix, Suffix

__all__ = ["parse_nodes_to_list", "SepBlockList"]


def create_BlockList(nodes, has_header, listclass=BlockList, config=None):
    """
    Simple wrapper to turn boolean `has_header` into a header `Node` argument
    to the `BlockList` class constructor (or any substituted `BlockList` class).
    """
    header = nodes.pop(0) if has_header else None
    # for now only set up to handle `SepBlockList` config
    listclasscheck = listclass is SepBlockList
    if config and listclass is SepBlockList:
        sep = config.get("sep") if "sep" in config else ":"
        hsep = config.get("headersep") if "headersep" in config else False
        lab = config.get("labels") if "labels" in config else None
        return listclass(nodes, header, sep=sep, headersep=hsep, labels=lab)
    else:
        return listclass(nodes, header)


def parse_nodes_to_list(nodes, listconfig=None, listclass=BlockList):
    """
    Generator function which yields all lists per block, ignoring any intervening
    or pre/succeeding non-list nodes (except any directly preceding 'header' node,
    with a line-terminating colon marked as `Suffix.ListInit`).
    """
    nodes = nodes.copy()  # do not modify input!
    list_open = list_header_open = False
    has_header = False
    parsed = []  # nodes belonging to a list go here
    while nodes:  # unlike a for loop, allows passing any remaining `nodes` upon return
        node = nodes.pop(0)
        if list_open:
            # very crudely, just terminate on any non-list continuation item
            if node.prefix is Prefix.ContList:
                parsed.append(node)
            else:
                list_open = False  # list terminates
                # a break here would skip any remaining nodes, requiring func recursion
                # to capture multiple lists per block, a yield is the preferable option
                yield create_BlockList(parsed, has_header, listclass, listconfig)
                has_header = False  # reset this in case parsing another list in block
                parsed = []
        elif list_header_open:
            if node.prefix == Prefix.InitList:
                list_header_open = False
                list_open = True
                parsed.append(node)
        else:
            if node.suffix == Suffix.InitList:
                list_header_open = has_header = True
                parsed.append(node)
            elif node.prefix == Prefix.InitList:
                if list_header_open:
                    list_header_open = False
                list_open = True
                parsed.append(node)
    if parsed:
        yield create_BlockList(parsed, has_header, listclass, listconfig)


def _as_df(self, forbid_header=False):
    labels = self._labels
    nodes = self.all_nodes if self.has_sep_header and not forbid_header else self.nodes
    datadict = {label: [getattr(node, label) for node in nodes] for label in labels}
    return DataFrame.from_dict(datadict)


class SepBlockList(BlockList):
    """
    A structured variant of the `BlockList` block-level list, with a
    specified separator, to function in the manner of a CSV,
    optionally with the 'columns' named via the `labels` argument.

    Will not permit different numbers of values in the header (if the
    header exists but does not conform to the CSV format, `sep_header`
    can disable the attempt to validate the header node in this way).

    `tokenise_separated_values` implements an interface for setting
    attributes on the `SepBlockList` according to a given list of
    attribute names (TODO: method to create a dictionary from the list
    nodes, to then e.g. create a `pandas.DataFrame`).
    """

    def __init__(self, nodes, h=None, par=None, sep=":", headersep=False, labels=None):
        super().__init__(nodes, header=h, parent_elem=par)
        self.sep = sep
        self.has_sep_header = headersep
        self.tokenise_separated_values(labels)  # sets `parts` attribute
        self._labels = labels
        if labels:
            self.as_df = _as_df.__get__(self)

    def tokenise_separated_values(self, sep_header_labels=None):
        n_parts = None
        if (
            self.has_sep_header
        ):  # there should be a header, to be parsed for sep. values
            assert self.header, "Cannot parse header: list is unheadered"
            header_offset = 0
            if sep_header_labels:
                n_labels = len(sep_header_labels)
                labsetcount = len(set(sep_header_labels))
                labsetcheck = labsetcount == labsetcount
                assert labsetcheck, f"{n_labels} labels =/= {labsetcount} values"
                assert "parts" not in sep_header_labels, "'parts' is a reserved label"
        elif self.header:  # there is a header but not parsing for sep. values
            header_offset = 1
        else:  # there is no header so nothing to skip
            header_offset = 0
        for node in self.all_nodes[header_offset:]:
            parts = node.contents.split(self.sep)
            if n_parts:  # if sep. values count per node already set on first seen node
                lencheck = len(parts) == n_parts
                assert lencheck, f"{node} has {len(parts)} parts (expected {n_parts})"
            else:
                n_parts = len(parts)
                if sep_header_labels:
                    n_labels = len(sep_header_labels)
                    labcountcheck = n_labels == n_parts
                    assert labcountcheck, f"{n_labels} labels =/= {n_parts} values"
            if sep_header_labels:
                for label, val in zip(sep_header_labels, parts):
                    setattr(node, label, val)
            setattr(node, "parts", parts)



from enum import Enum
from functools import partial
from itertools import chain
from os import walk
from pathlib import Path
from sys import stderr

from bs4 import Tag

from ...__share__ import alphanum2num_monthdict
from ...scan.io import mmd
from ..yaml_util import get_yaml_manifest
from .html_index_util import EmittedIndex, IntermedDirIndex, WireIndex
from .html_page_util import HtmlDoc, HtmlPage
from .html_util import Attrs, CustomHtmlTag, PartialAttrs
from .sinks import DocSink

__all__ = ["BaseEmitter", "Emitter"]


class BaseEmitter:
    """
    The minimum viable task of an emitter is to take a directory of MMD documents
    and transform it into a corresponding HTML output.

    The first built on top of this is `WireEmitter`, which uses a `wire.mmt` file
    to specify which files in the directory will be emitted (not necessary if all
    will be).
    """

    @property
    def doc_subpaths(self):
        "Probably doesn't work/don't need (to be deleted...)"
        return [x.src_filepath.relative_to(self.dir) for x in self.docs]


def match_date_parts(path_parts):
    for i, p in enumerate(path_parts):
        if i == 0 or i >= len(path_parts) - 1:
            continue  # skip first and last parts for indexing trios
        if p in alphanum2num_monthdict:
            p_pre, p_post = path_parts[i - 1], path_parts[i + 1]
            if p_pre.isnumeric() and p_post.isnumeric():
                numeric_month = str(alphanum2num_monthdict.get(p))
                return p_pre, numeric_month, p_post
    raise ValueError(f"No alphanumeric month-containing date found in '{path_parts}'")


class Wire:
    """
    `Wire` objects are created as `WireEmitter._Wires` in a listcomp
    within the method `WireEmitter.procure_documents`.
    """

    def __init__(self, filepath, verbose=False):
        self.date = match_date_parts(filepath.parts)
        self.source = mmd(filepath)
        self.source_dir = filepath.parent
        self.sinks = []
        for b in self.source.blocks:
            # TODO: add a lever parsing class rather than manually doing this
            d = dict([map(str.strip, n.contents.split("⠶")) for n in b.nodes])
            source_files = d.get("transmit").split(",")
            template = d.get("template")
            # Enum wrapper to template class: select the class from template string
            sink_func = getattr(DocSink, template).value
            # positional arg partial func: different files share same source directory
            make_sink = partial(sink_func, self.date, self.source_dir, verbose=verbose)
            templated_sinks = list(map(make_sink, source_files))
            self.sinks.extend(templated_sinks)


class WireEmitter(BaseEmitter):
    """
    The `.sink`s on the `._Wire`s are the `.docs`, emitted upon `._trigger()`.
    """

    def __init__(self, directory, name, verbose=False):
        self._verbose = verbose
        if self._verbose:
            print(f"== WireEmitter: '{name}'==", file=stderr)
        self.dir = directory
        subdir = get_yaml_manifest("poll").pages.script.cfg.get("subdirectory")  # site
        self.target_dir = (self.dir / "..").resolve() / subdir / "wire"
        self.name = name
        self._procure_documents()  # populate the docs property via wire files
        self.emit_log = []  # populated for the first time upon trigger
        self._trigger()
        if self._verbose:
            print(
                "Emit log: \n  " + "\n  ".join([str(x) for x in self.emit_log]),
                file=stderr,
            )
        self._index_emissions()

    def _procure_documents(self):
        walker = walk(self.dir)
        wire = "wire.mmt"
        dirs = [Path(p) for (p, _, filenames) in walker if wire in filenames]
        if self._verbose:
            print("TARGET DIR ::", self.target_dir, file=stderr)
            print("All input wire dirs:", file=stderr)
            for p in dirs:
                print("  " + str(p), file=stderr)
        self.wires = [(p / wire) for p in dirs]
        self._Wires = [Wire(w, verbose=self._verbose) for w in self.wires]
        self.docs = list(chain.from_iterable([w.sinks for w in self._Wires]))

    def _trigger(self):
        # put the files in the new directory
        emitted_paths = []
        for d in self.docs:
            # write output to target in the target directory
            emitted_path = d.emit(self.target_dir)
            emitted_paths.append(emitted_path)
        self.emit_log.extend(emitted_paths)  # set as empty list in __init__

    def _index_emissions(self):
        lambda_sort = lambda x: [
            int(str(p)) if str(p).isnumeric() else str(p) for p in x.parts
        ]
        all_emission_dirs = sorted(
            set([p.parent for p in self.emit_log]), key=lambda_sort
        )
        if self._verbose:
            print(
                "Emit dirs:\n  " + "\n  ".join([str(x) for x in all_emission_dirs]),
                file=stderr,
            )
        rel_dirs = [d.relative_to(self.target_dir) for d in all_emission_dirs]
        if self._verbose:
            print(
                "All wire emitted dirs:\n  " + "\n  ".join([f"{d}" for d in rel_dirs])
            )
        all_subdirs = [[d] + [x for x in d.parents if x.parts] for d in rel_dirs]
        unique_subdirs = []  # didn't feel like using a set
        for l in all_subdirs:
            for d in l:
                if d in unique_subdirs:
                    continue
                unique_subdirs.append(d)
        if self._verbose:
            print(
                "unique subdirs: \n  "
                + "\n  ".join([f"{s}" for s in sorted(unique_subdirs)])
            )
        su_subdirs = sorted(unique_subdirs)
        top_level_subdirs = [d for d in sorted(unique_subdirs) if len(d.parts) == 1]
        self._index_emission_dir(Path("."), top_level_subdirs)
        for d in su_subdirs:
            self._index_emission_dir(d, unique_subdirs)

    def _index_emission_dir(self, e_dir, all_subdirs):
        f_name = "index.html"
        full_e_dir = self.target_dir / e_dir
        p = full_e_dir / f_name
        if len(e_dir.parts) == 0:  # i.e. the top-level path "."
            files_to_log = []
        else:
            files_to_log = [e for e in self.emit_log if e.parent == full_e_dir]
            names_to_log = [p.relative_to(full_e_dir) for p in files_to_log]
        if self._verbose:
            print(f"  ⇢ index files: {e_dir}")
        if files_to_log:
            # i.e. if there are direct file descendants of the directory (not just subdirs)
            index_html_file = EmittedIndex(names_to_log, e_dir).as_str()
            n_f = len(files_to_log)
            f_str = f"{n_f} file{'s' if n_f != 1 else ''}"
            if self._verbose:
                print(f"    Writing index of {f_str} at {p}")
        else:
            if len(e_dir.parts) == 0:  # i.e. the top-level path "."
                subdirs = all_subdirs
                index_html_file = WireIndex(subdirs, e_dir).as_str()
            else:
                subdirs = [
                    d.relative_to(e_dir) for d in all_subdirs if d.parent == e_dir
                ]
                index_html_file = IntermedDirIndex(subdirs, e_dir).as_str()
            n_sd = len(subdirs)
            subdir_str = f"{n_sd} subdirector{'ies' if n_sd != 1 else 'y'}"
            if self._verbose:
                print(f"    Writing index of {subdir_str} at {p}")
        with open(p, "w") as f:
            f.write(index_html_file)
        return


class Emitter(Enum):
    wire = WireEmitter



__all__ = ["create_ns_aliases", "create_routing_table", "create_manifest"]


def create_ns_aliases(par):
    file_path = par / "alias.mmd"
    lines = [
        "-spin.systems=ss",
        "",
        "-spin=∫:",
        "-:log=",
        "-,:ocu=",
        "-,:arc=app",
        "-,:erg=",
        "-,:opt=",
        "-,:arb=",
        "-,:doc=",
        "",
        "-qu=q:",
        "-:cal=",
        "-,:conf=",
        "-,:pore=biorx",
        "-,:qrx=arx",
        "-,:poll=",
        "-,:reed=",
        "-,:noto=ruinoto",
        "-,:plot=plotspot",
        "-,:labs=",
    ]
    with open(file_path, "w") as f:
        f.write("\n".join(lines))


def create_routing_table(par):
    file_path = par / "routing.mmd"
    lines = [
        "-spin.systems=",
        "",
        "-spin:",
        "-:log=",
        "-,:ocu=",
        "-,:arc=",
        "-,:erg=",
        "-,:opt=",
        "-,:arb=",
        "-,:doc=",
        "",
        "-qu:",
        "-:cal=",
        "-,:conf=",
        "-,:pore=",
        "-,:qrx=",
        "-,:poll=",
        "-,:reed=",
        "-,:noto=",
        "-,:plot=",
        "-,:labs=",
    ]
    with open(file_path, "w") as f:
        f.write("\n".join(lines))


def create_manifest(par):
    file_path = par / "manifest.mmd"
    lines = [
        "-spin.systems:spin-systems:master www:",
        "-:cal:qu-cal:master www",
        "-,:log:spin-log:master www",
        "-,:conf:qu-conf:master www",
        "-,:pore:qu-pore:master www",
        "-,:ocu:naiveoculus:master www",
        "-,:arc:appendens:master www",
        "-,:qrx:qu-arx:master www",
        "-,:erg:spin-erg:master www",
        "-,:opt:spin-opt:master www",
        "-,:poll:qu-poll:master www",
        "-,:arb:spin-arb:master www",
        "-,:reed:qu-reed:master www",
        "-,:noto:qu-noto:master www",
        "-,:plot:qu-plot:master www",
        "-,:doc:spin-doc:master www",
        "-,:labs:qu-labs:master www",
    ]
    with open(file_path, "w") as f:
        f.write("\n".join(lines))



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    reqs = fh.read().splitlines()

def local_scheme(version):
    return ""

def version_scheme(version):
    return version.tag.base_version

setup(
    name="ql",
    author="Louis Maddox",
    author_email="louismmx@gmail.com",
    description="spin.systems generator and driver",
    license="MIT License",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/spin-systems/quill/",
    packages=find_packages("src"),
    package_dir={"": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
    ],
    include_package_data=True,
    use_scm_version={
        "write_to": "version.py",
        "version_scheme": version_scheme,
        "local_scheme": local_scheme,
    },
    setup_requires=["setuptools_scm"],
    install_requires=reqs,
    python_requires=">=3",
)



from bs4 import Tag

from .html_util import CustomHtmlTag

__all__ = ["NavLinkList", "UlLinkList", "MetaTag"]

###### Utility classes for specific tags


class NavLinkList(CustomHtmlTag):
    def __init__(self, titles, links, nav_params={}, ul_params={}, link_params={}):
        super().__init__(name="nav", **nav_params)
        ul = UlLinkList(titles, links, ul_params=ul_params, link_params=link_params)
        self.append(ul)


class UlLinkList(CustomHtmlTag):
    def __init__(self, titles, links, ul_params={}, link_params={}):
        self.titles = titles
        self.links = links
        super().__init__(name="ul", **ul_params)
        li_tags = []
        if "attrs" in link_params:
            extra_attrs = link_params.pop("attrs")
        else:
            extra_attrs = {}
        for t, l in zip(self.titles, self.links):
            li_tag = Tag(name="li")
            # note that here the href link is coerced to string automatically
            link_attrs = {"href": l, **extra_attrs}
            a_tag = Tag(name="a", attrs=link_attrs, **link_params)
            a_tag.append(t)  # the contents here can be string or a nested tag passed in
            li_tag.append(a_tag)
            self.append(li_tag)


class MetaTag(Tag):
    def __init__(self, attrs):
        super().__init__(name="meta", attrs=attrs, can_be_empty_element=True)


class NavHeader(CustomHtmlTag):
    def __init__(self, size=1, breadcrumbs=None, nh_params={}):
        default_nh_params = {"id": "nav_header"}
        # nh_params = {**nh_params, **self.nav_header_attrs} # merge with args if provided
        super().__init__(name=f"h{size}", **nh_params)
        if breadcrumbs:
            # construct it with the sep
            self.append(breadcrumbs)


class BreadCrumb(CustomHtmlTag):
    def __init__(self, text, link=None, tagname="a", link_params={}):
        link_attrs = {"class": "crumb"}
        if link:
            if "attrs" in link_params:
                extra_attrs = link_params.pop("attrs")
                link_attrs = {**link_attrs, **extra_attrs}
            link_attrs.update({"href": link})
        super().__init__(name=tagname, attrs=link_attrs, **link_params)
        self.append(text)



__all__ = ["parse_man_node", "read_man"]


def parse_man_node(node, host="gitlab.com", hosting_at="gitlab.io"):
    """
    Parse the ['human-authored'] sub/domain name and ['slugged'] repo name prefix
    from the manifest (a list in MMD file format). Apex and subdomain addresses
    are treated the same way by this function.
    """
    colsv = node.contents.split(":")[:2]
    assert (
        len(colsv) == 2
    ), f"Expected 2 colon-separated values, got {len(colsv)} ({node})"
    domain, repo = colsv
    org = repo
    ref = f"git@{host}:{org}/{repo}.{hosting_at}.git"
    return domain, ref


def read_man(man):
    """
    Process a `DocList` in `.mmd` list format, i.e. with colon-sep. values,
    as `subdomain:group` [where group doubles as the repo name prefix for
    GitLab] beginning with the apex domain as a list header, and followed by
    subdomains to be applied to the apex domain).
    """
    assert man.list, "The manifest MMD should be a single list"
    repo_info = list(map(parse_man_node, man.list.all_nodes))
    return repo_info


def read_man_df(man):
    repo_info = read_man(man)
    df = man.as_df()
    git_addr = [x[1] for x in repo_info]
    df["git_url"] = git_addr
    return df



from .lever import MMD

__all__ = ["mmd"]


def mmd(filepath, listparseconfig=None):
    with open(filepath) as f:
        mmd_lines = f.readlines()
    return MMD(mmd_lines, listparseconfig=listparseconfig)



from enum import Enum
from pathlib import Path
from sys import stderr

from ...scan.io import mmd
from .html_page_util import HtmlDoc

__all__ = ["QADocSink", "DocSink"]


class BaseDocSink:
    def __init__(self, date, src_dir, src_file, verbose=True):
        self._verbose = verbose
        self.date = date
        self.src_file = src_file
        self.src_filepath = (src_dir / src_file).resolve()

    qa_pair = False
    nest_qa_pair_as_summary = False

    @property
    def target(self):
        return Path(self.src_file).stem  # i.e. strip .mmd file extension

    def emit(self, target_dir):
        self.transform()
        # dumb approach, but do it for now
        date_subpath = Path(*self.date)
        emit_dir = target_dir / date_subpath
        if not emit_dir.exists():
            emit_dir.mkdir(parents=True)
        emit_path = emit_dir / self.target
        if self._verbose:
            print(f"Emitting --> {emit_path}", file=stderr)
        with open(emit_path, "w") as f:
            f.write(self.output)
        return emit_path

    def transform(self):
        src = mmd(self.src_filepath)
        listparseconfig = {"listclass": "auto", "strict_list_breaks": False}
        src = mmd(self.src_filepath, listparseconfig)
        file_depth = 4  # len(["wire", "yy", "mm", "dd"])
        self.html_doc = HtmlDoc(
            src,
            depth_from_root=file_depth,
            qa_pair=self.qa_pair,
            nested_qa_summary=self.nest_qa_pair_as_summary,
        )
        self.transform_hook()
        self.output = self.html_doc.as_str()

    def transform_hook(self):
        """
        Extensible hook for subclasses to intervene in `transform`.
        """
        pass


class QADocSink(BaseDocSink):
    """
    Shallow subclass of BaseDocSink but pairing up Q&A
    """

    qa_pair = True


class RadioTranscriptDocSink(BaseDocSink):
    """
    Unchanged from BaseDocSink but will probably modify to suit radio.
    """

    pass


class RadioTranscriptSummariesDocSink(RadioTranscriptDocSink):
    """
    Summary version of `RadioTranscriptDocSink`, Q&A pairs are transformed
    to become summary and details elements.
    """

    qa_pair = True
    nest_qa_pair_as_summary = True  # Q & A will become summary & details
    # def transform_hook(self):
    #    # Manually surround pairs of nodes as summary/details
    #    qa_pair_idx_dict = {}
    #    for i, b in enumerate(self.html_doc.doc.blocks):
    #        qa_pair_idx_block_dict = {}
    #        for j, n in enumerate(b.nodes):
    #            if hasattr(n, "paired_to"):
    #                aired_bl_no, paired_line_no = n.paired_to
    #                qa_pair_idx_block_dict[j] = paired_line_no
    #        qa_pair_idx_dict[i] = qa_pair_idx_block_dict
    #    #print(qa_pair_idx_dict)


class DocSink(Enum):
    qa = QADocSink
    speech = QADocSink  # TODO: change
    slides = QADocSink  # TODO: change
    roundtable = QADocSink  # TODO: change
    radio_transcript = RadioTranscriptDocSink
    radio_transcript_summaries = RadioTranscriptSummariesDocSink



from aenum import NamedConstant

from ...__share__ import classproperty

__all__ = ["tokenise_line", "Prefix", "Suffix", "Node"]


def has_precedent(seen, prefix):
    """
    Check if there is a seen node with a given prefix.
    No need to do this for suffixes as their precedent only ever checked
    for directly preceding.
    """
    return any([n.prefix is prefix for n in seen])


def lever_config_dict():
    """
    Return default parser config.
    """
    return {
        # Permit `Prefix.ListInit` without directly preceding `Suffix.ListInit`
        "ALLOW_LIST_WITHOUT_SUFFIX_INIT": True,
    }


def tokenise_line(line, line_no, block_no, seen=None, config=None):
    """
    Follow rules for prefix and suffix token order, splitting
    the string into `(prefix, contents, suffix)`. Optionally
    specify `custom_config` dict to override defaults from
    `lever_config_dict`.
    """
    # Begin config instantation
    default_config = lever_config_dict()
    if config:
        # Only allow custom config keys also present in default_config dict
        config = {k: v for k, v in config.items() if k in default_config}
        config = {**default_config, **config}  # override default config
    else:
        config = default_config
    # Read config variables
    # TODO: just read into locals if config gets any more complicated
    ALLOW_LIST_WITHOUT_SUFFIX_INIT = config.get("ALLOW_LIST_WITHOUT_SUFFIX_INIT")
    # End config instantation
    seen = seen if seen else []  # avoid default mutable arg gotcha
    penultimate = seen[-1] if seen else None  # directly preceding
    node = Node(
        prefix=None, contents=line, suffix=None, line_no=line_no, block_no=block_no
    )
    if line == "":
        node.prefix = Prefix.BlankNode
    elif line.startswith("-:"):
        if line[2] == "'":
            node.prefix = Prefix.Because
        elif line[2] == ".":
            node.prefix = Prefix.Therefore
        elif penultimate:
            if penultimate.prefix is Prefix.Question:
                node.prefix = Prefix.Answer
                # Keep a record of the location of its paired node
                penultimate.paired_to = (block_no, line_no)
                node.paired_to = (penultimate.block_no, penultimate.line_no)
            elif penultimate.contents.endswith(":"):
                # Mark penultimate line's suffix as beginning a list
                penultimate.suffix = Suffix.InitList
                node.prefix = Prefix.InitList
            elif ALLOW_LIST_WITHOUT_SUFFIX_INIT:
                node.prefix = Prefix.InitList
            else:
                print(f"Falling through... {node.contents}")
        elif ALLOW_LIST_WITHOUT_SUFFIX_INIT:
            node.prefix = Prefix.InitList
        # (Question not implemented here)
    elif line.startswith("-?"):
        node.prefix = Prefix.Question
    elif line.startswith("-,"):
        if line[2] == ":" and has_precedent(seen, Prefix.InitList):
            # For now consider a list 'open' indefinitely while the block is
            # i.e. until blank line (TODO implement break at section divider)
            node.prefix = Prefix.ContList
        elif line[2] == ",":
            node.prefix = Prefix.Ascent
        elif line[2] == "?":
            node.prefix = Prefix.ContQuestion
        else:
            node.prefix = Prefix.FollowOn
    elif line.startswith("-."):
        if line[2] == ".":
            node.prefix = Prefix.Descent
        elif line[2] == ",":
            node.prefix = Prefix.ContDesc
    elif line.startswith("-#"):
        max_header = 8  # 1-based count of header levels
        start_at = len(Prefix.Header1.str)
        confirmed_levels = 1
        for i, l in enumerate(line[start_at : start_at + max_header]):
            if l != "#":
                break
            else:
                confirmed_levels += 1
        if confirmed_levels > max_header:
            raise ValueError(f"Can't parse header beyond level {max_header} ('{line}')")
        header_name = f"Header{confirmed_levels}"
        node.prefix = Prefix.token_from_name(header_name)
    elif line.startswith("-~==~-"):
        node.prefix = Prefix.SectBreak
    elif line.startswith("-"):
        node.prefix = Prefix.PlainNode
    else:
        raise ValueError(f"Unable to tokenise line: '{line}'")
    # If this succeeded without raising an error, make contents
    return node


class Affix(NamedConstant):
    """
    Affixes following the MMD 'lever' format specification.
    """

    @classmethod
    def name_from_tokenseq(cls, tok):
        """
        IN: the string matched by the token. OUT: the node name or
        a list of names if the token name is not unique.
        """
        matched = cls.tokseq2name_multidict.get(tok)
        if matched:
            if len(matched) == 1:
                [matched] = matched
        else:
            raise KeyError(f"{tok} not a valid Affix token sequence")
        return matched

    # @classproperty
    # def tokseq2name_dict(cls):
    #    "KEY: the string matched by the token. VAL: the node name."
    #    # N.B. does not assume the token-matching string is unique
    #    d = {k: v._value_ for k,v in Prefix.__members__.items()}
    #    return d

    @classproperty
    def tokseq2name_multidict(cls):
        "KEY: the string matched by the token(s). VAL: list of node name(s)."
        # N.B. does not assume the token-matching string is unique
        d = {
            v._value_: [
                k for k in cls._members_ if cls._members_.get(k)._value_ == v._value_
            ]
            for v in cls._members_.values()
        }
        return d

    @classmethod
    def token_from_name(cls, name):
        "IN: the node name. OUT: the token (a `NamedConstant` object)."
        return cls.name2tok_dict.get(name)

    @classproperty
    def name2tok_dict(cls):
        "KEY: the node name. VAL: the token (a `NamedConstant` object)."
        d = {v._name_: v for v in cls._members_.values()}
        return d

    @property
    def str(self):
        return self._value_

    def is_named(self, name):
        return self._name_ == name


class Prefix(Affix):
    """
    Prefixes following the MMD 'lever' format specification. Capital
    letters indicate token symbols [A to F, as well as G_ ('gap')].
    Headings levels 1-8 are named Ha through Hh. Empty lines are named Z_.
    Note that a blank line and a section break are given as a prefix although
    they each span the entire line (underscore indicates a 'terminating prefix').
    """

    PlainNode = "-"  # A or Cb [if line ends with colon and precedes C"
    FollowOn = "-,"  # Ac
    InitList = "-:"  # B [default: indicating a list]
    Answer = "-:"  # Cr [if penultimate line was C]
    ContList = "-,:"  # Bc
    Question = "-?"  # C
    ContQuestion = "-,?"  # Cc
    Descent = "-.."  # D
    ContDesc = "-.,"  # Dc
    Ascent = "-,,"  # Du
    Therefore = "-:."  # E
    Because = "-:'"  # F
    SectBreak = "-~==~-"  # G_
    Header1 = "-#"  # Ha
    Header2 = "-##"  # Hb
    Header3 = "-###"  # Hc
    Header4 = "-####"  # Hd
    Header5 = "-#####"  # He
    Header6 = "-######"  # Hf
    Header7 = "-#######"  # Hg
    Header8 = "-########"  # Hh
    BlankNode = ""  # Z_


class Suffix(Affix):
    """
    Suffixes following the MMD 'lever' format specification'.
    """

    InitList = ":"


class Node:
    """
    Provide a simple string-repr while storing the Prefix and Suffix
    as properties (intervening `contents` is just a string for now).
    """

    def __init__(
        self, prefix=None, contents=None, suffix=None, line_no=None, block_no=None
    ):
        self.prefix = prefix
        self.contents = contents
        self.suffix = suffix
        self.line_no = line_no
        self.block_no = block_no

    def __repr__(self):
        pre = self.prefix.str if self.prefix else ""
        con = self.contents if self.contents else ""
        suf = self.suffix.str if self.suffix else ""
        return f"{pre}{con}{suf}"

    @property
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self, p):
        if hasattr(self, "_prefix"):
            # modifying an already set prefix so must also modify `self.contents`
            ex_prefix = self.prefix
            if ex_prefix:
                assert self.contents, "Tried to change prefix on contents-less Node"
                self.contents = ex_prefix.str + self.contents  # reconstitute
            if p:
                self.contents = self.contents[len(p.str) :]
        if p:
            assert isinstance(p, Prefix)  # either `None` or `Prefix` NamedConstant
        self._prefix = p

    @property
    def contents(self):
        return self._contents

    @contents.setter
    def contents(self, c):
        self._contents = c

    @property
    def suffix(self):
        return self._suffix

    @suffix.setter
    def suffix(self, s):
        if hasattr(self, "_suffix"):
            # modifying an already set suffix so must also modify `self.contents`
            ex_suffix = self.suffix
            if ex_suffix:
                assert self.contents, "Tried to change suffix on content-less Node"
                self.contents = self.contents + ex_suffix.str
            if s:
                self.contents = self.contents[: -len(s.str)]
        if s:
            assert isinstance(s, Suffix)  # either `None` or `Suffix` NamedConstant
        self._suffix = s
